from odak import np
from odak.tools import distance_between_two_points

def propagate_field(points0,points1,field0,wave_number,direction=1):
    """
    Definition to propagate a field from points to an another points in space.

    Parameters
    ----------
    points0       : ndarray
                    Start points (i.e. odak.tools.grid_sample).
    points1       : ndarray
                    End points (ie. odak.tools.grid_sample).
    field0        : ndarray
                    Field for given starting points.
    wave_number   : float
                    Wave number of a wave, see odak.wave.wavenumber for more.
    direction     : float
                    For propagating in forward direction set as 1, otherwise -1.

    Returns
    ----------
    field1        : ndarray
                    Field for given end points.
    """
    field1 = np.zeros(points1.shape[0],dtype=np.complex64)
    for point_id in range(points0.shape[0]):
        point     = points0[point_id]
        distances = distance_between_two_points(
                                                point,
                                                points1
                                               )
        field1   += electric_field_per_plane_wave(
                                                  calculate_amplitude(field0[point_id]),
                                                  distances*direction,
                                                  wave_number,
                                                  phase=calculate_phase(field0[point_id])
                                                 )
    return field1

def propagate_plane_waves(field,opd,k,w=0,t=0):
    """
    Definition to propagate a field representing a plane wave at a particular distance and time.

    Parameters
    ----------
    field        : complex
                   Complex field.
    opd          : float
                   Optical path difference in mm.
    k            : float
                   Wave number of a wave, see odak.wave.parameters.wavenumber for more.
    w            : float
                   Rotation speed of a wave, see odak.wave.parameters.rotationspeed for more.
    t            : float
                   Time in seconds.

    Returns
    ----------
    new_field     : complex
                    A complex number that provides the resultant field in the complex form A*e^(j(wt+phi)).
    """
    new_field = field*np.exp(1j*(-w*t+opd*k))/opd**2
    return new_field

def electric_field_per_plane_wave(amplitude,opd,k,phase=0,w=0,t=0):
    """
    Definition to return state of a plane wave at a particular distance and time.

    Parameters
    ----------
    amplitude    : float
                   Amplitude of a wave.
    opd          : float
                   Optical path difference in mm.
    k            : float
                   Wave number of a wave, see odak.wave.parameters.wavenumber for more.
    phase        : float
                   Initial phase of a wave.
    w            : float
                   Rotation speed of a wave, see odak.wave.parameters.rotationspeed for more.
    t            : float
                   Time in seconds.

    Returns
    ----------
    field        : complex
                   A complex number that provides the resultant field in the complex form A*e^(j(wt+phi)).
    """
    field = amplitude*np.exp(1j*(-w*t+opd*k+phase))/opd**2
    return field

def calculate_phase(field,deg=False):
    """
    Definition to calculate phase of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.
    deg          : bool
                   If set True, the angles will be returned in degrees.

    Returns
    ----------
    phase        : float
                   Phase or phases of electric field(s) in radians.
    """
    phase = np.angle(field)
    if deg == True:
        phase *= 180./np.pi
    return phase

def calculate_amplitude(field):
    """
    Definition to calculate amplitude of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.

    Returns
    ----------
    amplitude    : float
                   Amplitude or amplitudes of electric field(s).
    """
    amplitude = np.abs(field)
    return amplitude

def calculate_intensity(field):
    """
    Definition to calculate intensity of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.

    Returns
    ----------
    intensity    : float
                   Intensity or intensities of electric field(s).
    """
    intensity = np.abs(field)**2
    return intensity

def wavenumber(wavelength):
    """
    Definition for calculating the wavenumber of a plane wave.

    Parameters
    ----------
    wavelength   : float
                   Wavelength of a wave in mm.

    Returns
    ----------
    k            : float
                   Wave number for a given wavelength.
    """
    k = 2*np.pi/wavelength
    return k

def rotationspeed(wavelength,c=3*10**11):
    """
    Definition for calculating rotation speed of a wave (w in A*e^(j(wt+phi))).

    Parameters
    ----------
    wavelength   : float
                   Wavelength of a wave in mm.
    c            : float
                   Speed of wave in mm/seconds. Default is the speed of light in the void!

    Returns
    ----------
    w            : float

    """
    f = c*wavelength
    w = 2*np.pi*f
    return w




from odak import np
import torch

def quadratic_phase_function(nx,ny,k,focal=0.4,dx=0.001,offset=[0,0]):
    """ 
    A definition to generate 2D quadratic phase function, which is typically use to represent lenses.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    focal      : float
                 Focal length of the quadratic phase function.
    dx         : float
                 Pixel pitch.
    offset     : list
                 Deviation from the center along X and Y axes.

    Returns
    ---------
    function   : torch.tensor
                 Generated quadratic phase function.
    """
    size  = [ny,nx]
    x     = torch.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])-offset[1]*dx
    y     = torch.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])-offset[0]*dx
    X,Y   = torch.meshgrid(x,y)
    Z     = X**2+Y**2
    focal = torch.tensor([focal])
    k     = torch.tensor([k])
    qwf   = torch.exp(1j*k*0.5*torch.sin(Z/focal))
    return qwf

def prism_phase_function(nx,ny,k,angle,dx=0.001,axis='x'):
    """
    A definition to generate 2D phase function that represents a prism. See Goodman's Introduction to Fourier Optics book for more.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    angle      : float
                 Tilt angle of the prism in degrees.
    dx         : float
                 Pixel pitch.
    axis       : str
                 Axis of the prism.

    Returns
    ----------
    prism      : torch.tensor
                 Generated phase function for a prism.
    """
    angle = torch.deg2rad(torch.tensor([angle]))
    size  = [ny,nx]
    x     = torch.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])
    y     = torch.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])
    X,Y   = torch.meshgrid(x,y)
    if axis == 'y':
        prism = torch.exp(-1j*k*torch.sin(angle)*Y)
    elif axis == 'x':
        prism = torch.exp(-1j*k*torch.sin(angle)*X)
    return prism

def linear_grating(nx,ny,every=2,add=3.14,axis='x'):
    """
    A definition to generate a linear grating.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    every      : int
                 Add the add value at every given number.
    add        : float
                 Angle to be added.
    axis       : string
                 Axis eiter X,Y or both.

    Returns
    ----------
    field      : torch.tensor
                 Linear grating term.
    """
    grating = torch.zeros((nx,ny),dtype=torch.complex64)
    if axis == 'x':
        grating[::every,:] = torch.exp(torch.tensor(1j*add))
    if axis == 'y':
        grating[:,::every] = torch.exp(torch.tensor(1j*add))
    if axis == 'xy':
        checker  = np.indices((nx,ny)).sum(axis=0) % every
        checker  = torch.from_numpy(checker)
        checker += 1
        checker  = checker % 2
        grating  = torch.exp(1j*checker*add)
    return grating



import sys

def test():
    from odak import np
    from odak.wave import gerchberg_saxton_3d,adjust_phase_only_slm_range,produce_phase_only_slm_pattern,calculate_amplitude,wavenumber,propagate_beam
    from odak.tools import save_image
    wavelength                   = 0.000000532
    dx                           = 0.0000064
    distances                    = np.array([0.2,0.1])
    input_fields                 = np.random.rand(2,500,500).astype(np.complex64)
    iteration_number             = 100
    distance_light_slm           = 2.0
    k                            = wavenumber(wavelength)
    hologram                     = gerchberg_saxton_3d(
                                                       input_fields,
                                                       iteration_number,
                                                       -distances,
                                                       dx,
                                                       wavelength,
                                                       np.pi*2,
                                                       'Bandlimited Angular Spectrum',
                                                       initial_phase=None
                                                      )
    reconstruction_0             = propagate_beam(hologram,k,distances[0],dx,wavelength,propagation_type='Bandlimited Angular Spectrum')
    reconstruction_1             = propagate_beam(hologram,k,distances[1],dx,wavelength,propagation_type='Bandlimited Angular Spectrum')
    amplitude_0                  = calculate_amplitude(reconstruction_0)
    amplitude_1                  = calculate_amplitude(reconstruction_1)
    hologram,_                   = produce_phase_only_slm_pattern(
                                                                  hologram,
                                                                  2*np.pi,
                                                                  'output_hologram.png'
                                                                 )
 
    save_image(
               'output_amplitude_0.png',
               amplitude_0,
               cmin=0,
               cmax=np.amax(amplitude_0)
              )
    save_image(
               'output_amplitude_1.png',
               amplitude_1,
               cmin=0,
               cmax=np.amax(amplitude_1)
              )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



from plyfile import PlyData, PlyElement
from odak import np
from .transformation import rotate_point

def read_PLY(fn,offset=[0,0,0],angles=[0.,0.,0.],mode='XYZ'):
    """
    Definition to read a PLY file and extract meshes from a given PLY file. Note that rotation is always with respect to 0,0,0.

    Parameters
    ----------
    fn           : string
                   Filename of a PLY file.
    offset       : ndarray
                   Offset in X,Y,Z.
    angles       : list
                   Rotation angles in degrees.
    mode         : str
                   Rotation mode determines ordering of the rotations at each axis. There are XYZ,YXZ,ZXY and ZYX modes. 

    Returns
    ----------
    triangles    : ndarray
                  Triangles from a given PLY file. Note that the triangles coming out of this function isn't always structured in the right order and with the size of (MxN)x3. You can use numpy's reshape to restructure it to mxnx3 if you know what you are doing.
    """
    if np.__name__ != 'numpy':
        import numpy as np_ply
    else:
        np_ply = np
    with open(fn,'rb') as f:
        plydata = PlyData.read(f)
    triangle_ids = np_ply.vstack(plydata['face'].data['vertex_indices'])
    triangles    = []
    for vertex_ids in triangle_ids:
        triangle     = [
                        rotate_point(plydata['vertex'][int(vertex_ids[0])].tolist(),angles=angles,offset=offset)[0],
                        rotate_point(plydata['vertex'][int(vertex_ids[1])].tolist(),angles=angles,offset=offset)[0],
                        rotate_point(plydata['vertex'][int(vertex_ids[2])].tolist(),angles=angles,offset=offset)[0]
                       ]
        triangle     = np_ply.asarray(triangle)
        triangles.append(triangle)
    triangles = np_ply.array(triangles)
    triangles = np.asarray(triangles,dtype=np.float)
    return triangles

def read_PLY_point_cloud(filename):
    """
    Definition to read a PLY file as a point cloud.

    Parameters
    ----------
    filename     : str
                   Filename of a PLY file.

    Returns
    ----------
    point_cloud  : ndarray
                   An array filled with poitns from the PLY file.
    """
    plydata      = PlyData.read(filename)
    if np.__name__ != 'numpy':
        import numpy as np_ply
        point_cloud      = np_ply.zeros((plydata['vertex'][:].shape[0],3))
        point_cloud[:,0] = np_ply.asarray(plydata['vertex']['x'][:])
        point_cloud[:,1] = np_ply.asarray(plydata['vertex']['y'][:])
        point_cloud[:,2] = np_ply.asarray(plydata['vertex']['z'][:])
        point_cloud      = np.asarray(point_cloud)
    else:
        point_cloud      = np.zeros((plydata['vertex'][:].shape[0],3))
        point_cloud[:,0] = np.asarray(plydata['vertex']['x'][:])
        point_cloud[:,1] = np.asarray(plydata['vertex']['y'][:])
        point_cloud[:,2] = np.asarray(plydata['vertex']['z'][:])
    return point_cloud

def write_PLY(triangles,savefn='output.ply'):
    """
    Definition to generate a PLY file from given points.

    Parameters
    ----------
    triangles   : ndarray
                  List of triangles with the size of Mx3x3.
    savefn      : string
                  Filename for a PLY file.
    """
    tris  = []
    pnts  = []
    color = [255,255,255]
    for tri_id in range(triangles.shape[0]):
       tris.append(
                   (
                    [3*tri_id,3*tri_id+1,3*tri_id+2], 
                    color[0],
                    color[1],
                    color[2]
                   )
                  )
       for i in range(0,3):
           pnts.append(
                       (
                        float(triangles[tri_id][i][0]),
                        float(triangles[tri_id][i][1]),
                        float(triangles[tri_id][i][2])
                       )
                      )
    if np.__name__ == 'cupy':
        import numpy as np_cpu
        tris   = np_cpu.asarray(tris, dtype=[('vertex_indices', 'i4', (3,)),('red', 'u1'), ('green', 'u1'),('blue', 'u1')])
        pnts   = np_cpu.asarray(pnts, dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')])
    else:
        tris   = np.asarray(tris, dtype=[('vertex_indices', 'i4', (3,)),('red', 'u1'), ('green', 'u1'),('blue', 'u1')])
        pnts   = np.asarray(pnts, dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')])
    # Save mesh.
    el1       = PlyElement.describe(pnts, 'vertex', comments=['Vertex data'])
    el2       = PlyElement.describe(tris, 'face', comments=['Face data'])
    PlyData([el1,el2],text="True").write(savefn)

def write_PLY_from_points(points,savefn='output.ply'):
    """
    Definition to generate a PLY file from given points.

    Parameters
    ----------
    points      : ndarray
                  List of points with the size of MxNx3.
    savefn      : string
                  Filename for a PLY file.

    """
    if np.__name__ != 'numpy':
        import numpy as np_ply
    else:
        np_ply = np
    # Generate equation
    samples = [points.shape[0],points.shape[1]]
    # Generate vertices.
    pnts    = []
    tris    = []
    for idx in range(0,samples[0]):
        for idy in range(0,samples[1]):
            pnt  = (points[idx,idy,0],points[idx,idy,1],points[idx,idy,2])
            pnts.append(pnt)
    color = [255,255,255]
    for idx in range(0,samples[0]-1):
        for idy in range(0,samples[1]-1):
            tris.append(([idy+(idx+1)*samples[0], idy+idx*samples[0]  , idy+1+idx*samples[0]], color[0], color[1], color[2]))
            tris.append(([idy+(idx+1)*samples[0], idy+1+idx*samples[0], idy+1+(idx+1)*samples[0]], color[0], color[1], color[2]))
    tris    = np_ply.asarray(tris, dtype=[('vertex_indices', 'i4', (3,)),('red', 'u1'), ('green', 'u1'),('blue', 'u1')])
    pnts    = np_ply.asarray(pnts, dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')])
    # Save mesh.
    el1     = PlyElement.describe(pnts, 'vertex', comments=['Vertex data'])
    el2     = PlyElement.describe(tris, 'face', comments=['Face data'])
    PlyData([el1,el2],text="True").write(savefn)



#!/usr/bin/env python

import sys

def test_plano_convex():
    import odak.raytracing as raytracer
    import odak.tools as tools
    import odak.catalog as catalog
    end_points        = tools.grid_sample(
                                          no=[5,5],
                                          size=[2.0,2.0],
                                          center=[0.,0.,0.],
                                          angles=[0.,0.,0.]
                                         )
    start_point       = [0.,0.,-5.]
    rays              = raytracer.create_ray_from_two_points(
                                                             start_point,
                                                             end_points
                                                            )
    lens              = catalog.plano_convex_lens()
    normals,distances = lens.intersect(rays)
    return True

    from odak import np
    import plotly
    import plotly.graph_objs as go    
    if np.__name__ == 'cupy':
        df = np.asnumpy(normals[:,0])
        dx = np.asnumpy(end_points)
        dy = np.asnumpy(start_point)
    else:
        df = normals[:,0]
        dx = end_points
        dy = start_point
    trace0 = go.Scatter3d(x=df[:,0],y=df[:,1],z=df[:,2])
    trace1 = go.Scatter3d(x=dx[:,0],y=dx[:,1],z=dx[:,2])
    trace2 = go.Scatter3d(x=[dy[0],],y=[dy[1],],z=[dy[2],])
    data   = [trace0,trace1,trace2]
    fig    = dict(data=data)
    plotly.offline.plot(fig)
    assert True==True

if __name__ == '__main__':
    sys.exit(test_plano_convex())



import sys
import os
import odak
from odak import np

def main():
    n           = [100,100]
    ranges      = [[500.,3000.],[10.,100.]]
    x,y         = np.mgrid[0:n[0],0:n[1]]
    focals      = x*(ranges[0][1]-ranges[0][0])/n[0]+ranges[0][0]
    apertures   = y*(ranges[1][1]-ranges[1][0])/n[1]+ranges[1][0]
    wavelength  = 0.0005
    resolutions = np.zeros((n[0],n[1],3))
    for i in range(0,n[0]):
        for j in range(0,n[1]):
            resolutions[i,j,0] = focals[i,j]
            resolutions[i,j,1] = apertures[i,j]
            resolutions[i,j,2] = odak.wave.rayleigh_resolution(
                                                               diameter=resolutions[i,j,1],
                                                               focal=resolutions[i,j,0],
                                                               wavelength=wavelength
                                                              )*1000 # Conversion to um.
    figure = odak.visualize.surfaceshow(
                                        title='Spatial resolution',
                                        labels=[
                                                'Throw distance (mm)',
                                                'Aperture size (mm) ',
                                                'Spatial resolution (um)'
                                               ],
                                        types=[
                                               'log',
                                               'log',
                                               'log'
                                              ],
                                        font_size=16,
                                        tick_no=[2,2,4],
                                       )
    figure.add_surface(
                       data_x=resolutions[:,:,0],
                       data_y=resolutions[:,:,1],
                       data_z=resolutions[:,:,2],
                       contour=False
                      )
    figure.show()
    assert True==True

if __name__ == '__main__':
    sys.exit(main())



from odak import np
from .transformation import rotate_points,rotate_point
from odak.raytracing import create_ray_from_two_points

def random_sample_point_cloud(point_cloud,no,p=None):
    """
    Definition to pull a subset of points from a point cloud with a given probability.

    Parameters
    ----------
    point_cloud  : ndarray
                   Point cloud array.
    no           : list
                   Number of samples.
    p            : list
                   Probability list in the same size as no.

    Returns
    ----------
    subset       : ndarray
                   Subset of the given point cloud.
    """
    choice = np.random.choice(point_cloud.shape[0],no,p)
    subset = point_cloud[choice,:]
    return subset 


def sphere_sample(no=[10,10],radius=1.,center=[0.,0.,0.],k=[1,2]):
    """
    Definition to generate a regular sample set on the surface of a sphere using polar coordinates.

    Parameters
    ----------
    no          : list
                  Number of samples.
    radius      : float
                  Radius of a sphere.
    center      : list
                  Center of a sphere.
    k           : list
                  Multipliers for gathering samples. If you set k=[1,2] it will draw samples from a perfect sphere.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples        = np.zeros((no[0],no[1],3))
    psi,teta       = np.mgrid[0:no[0],0:no[1]]
    psi            = k[0]*np.pi/no[0]*psi
    teta           = k[1]*np.pi/no[1]*teta
    samples[:,:,0] = center[0]+radius*np.sin(psi)*np.cos(teta)
    samples[:,:,1] = center[0]+radius*np.sin(psi)*np.sin(teta)
    samples[:,:,2] = center[0]+radius*np.cos(psi)
    samples        = samples.reshape((no[0]*no[1],3))
    return samples

def sphere_sample_uniform(no=[10,10],radius=1.,center=[0.,0.,0.],k=[1,2]):
    """
    Definition to generate an uniform sample set on the surface of a sphere using polar coordinates.

    Parameters
    ----------
    no          : list
                  Number of samples.
    radius      : float
                  Radius of a sphere.
    center      : list
                  Center of a sphere.
    k           : list
                  Multipliers for gathering samples. If you set k=[1,2] it will draw samples from a perfect sphere.


    Returns
    ----------
    samples     : ndarray
                  Samples generated.

    """
    samples        = np.zeros((no[0],no[1],3))
    row            = np.arange(0,no[0])
    psi,teta       = np.mgrid[0:no[0],0:no[1]]
    for psi_id in range(0,no[0]):
        psi[psi_id] = np.roll(row,psi_id,axis=0)
        teta[psi_id] = np.roll(row,-psi_id,axis=0)
    psi            = k[0]*np.pi/no[0]*psi
    teta           = k[1]*np.pi/no[1]*teta
    samples[:,:,0] = center[0]+radius*np.sin(psi)*np.cos(teta)
    samples[:,:,1] = center[1]+radius*np.sin(psi)*np.sin(teta)
    samples[:,:,2] = center[2]+radius*np.cos(psi)
    samples        = samples.reshape((no[0]*no[1],3))
    return samples

def box_volume_sample(no=[10,10,10],size=[100.,100.,100.],center=[0.,0.,0.],angles=[0.,0.,0.]):
    """
    Definition to generate samples in a box volume.

    Parameters
    ----------
    no          : list
                  Number of samples.
    size        : list
                  Physical size of the volume.
    center      : list
                  Center location of the volume.
    angles      : list
                  Tilt of the volume.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples          = np.zeros((no[0],no[1],no[2],3))
    x,y,z            = np.mgrid[0:no[0],0:no[1],0:no[2]]
    step             = [
                        size[0]/no[0],
                        size[1]/no[1],
                        size[2]/no[2]
                       ]
    samples[:,:,:,0] = x*step[0]+step[0]/2.-size[0]/2.
    samples[:,:,:,1] = y*step[1]+step[1]/2.-size[1]/2.
    samples[:,:,:,2] = z*step[2]+step[2]/2.-size[2]/2.
    samples          = samples.reshape((samples.shape[0]*samples.shape[1]*samples.shape[2],samples.shape[3]))
    samples          = rotate_points(samples,angles=angles,offset=center)
    return samples

def circular_sample(no=[10,10],radius=10.,center=[0.,0.,0.],angles=[0.,0.,0.]):
    """
    Definition to generate samples inside a circle over a surface.

    Parameters
    ----------
    no          : list
                  Number of samples.
    radius      : float
                  Radius of the circle.
    center      : list
                  Center location of the surface.
    angles      : list
                  Tilt of the surface.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples         = np.zeros((no[0]+1,no[1]+1,3))
    r_angles,r      = np.mgrid[0:no[0]+1,0:no[1]+1]
    r               = r/np.amax(r)*radius
    r_angles        = r_angles/np.amax(r_angles)*np.pi*2
    samples[:,:,0]  = r*np.cos(r_angles)
    samples[:,:,1]  = r*np.sin(r_angles)
    samples         = samples[1:no[0]+1,1:no[1]+1,:]
    samples         = samples.reshape((samples.shape[0]*samples.shape[1],samples.shape[2]))
    samples         = rotate_points(samples,angles=angles,offset=center)
    return samples

def circular_uniform_random_sample(no=[10,50],radius=10.,center=[0.,0.,0.],angles=[0.,0.,0.]):
    """ 
    Definition to generate sample inside a circle uniformly but randomly.

    Parameters
    ----------
    no          : list
                  Number of samples.
    radius      : float
                  Radius of the circle.
    center      : list
                  Center location of the surface.
    angles      : list
                  Tilt of the surface.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples = np.empty((0,3))
    rs      = np.sqrt(np.random.uniform(0,1,no[0]))
    angs    = np.random.uniform(0,2*np.pi,no[1])
    for i in rs:
        for angle in angs:
            r       = radius*i
            point   = np.array([float(r*np.cos(angle)),float(r*np.sin(angle)),0])
            samples = np.vstack((samples,point))
    samples = rotate_points(samples,angles=angles,offset=center)
    return samples

def circular_uniform_sample(no=[10,50],radius=10.,center=[0.,0.,0.],angles=[0.,0.,0.]):
    """
    Definition to generate sample inside a circle uniformly.

    Parameters
    ----------
    no          : list
                  Number of samples.
    radius      : float
                  Radius of the circle.
    center      : list
                  Center location of the surface.
    angles      : list
                  Tilt of the surface.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples = np.empty((0,3))
    for i in range(0,no[0]):
        r = i/no[0]*radius
        ang_no = no[1]*i/no[0]
        for j in range(0,int(no[1]*i/no[0])):
            angle   = j/ang_no*2*np.pi
            point   = np.array([float(r*np.cos(angle)),float(r*np.sin(angle)),0])
            samples = np.vstack((samples,point))
    samples = rotate_points(samples,angles=angles,offset=center)
    return samples

def grid_sample(no=[10,10],size=[100.,100.],center=[0.,0.,0.],angles=[0.,0.,0.]):
    """
    Definition to generate samples over a surface.

    Parameters
    ----------
    no          : list
                  Number of samples.
    size        : list
                  Physical size of the surface.
    center      : list
                  Center location of the surface.
    angles      : list
                  Tilt of the surface.

    Returns
    ----------
    samples     : ndarray
                  Samples generated.
    """
    samples        = np.zeros((no[0],no[1],3))
    step           = [
                      size[0]/(no[0]-1),
                      size[1]/(no[1]-1)
                     ]
    x,y            = np.mgrid[0:no[0],0:no[1]]
    samples[:,:,0] = x*step[0]-size[0]/2.
    samples[:,:,1] = y*step[1]-size[1]/2.
    samples        = samples.reshape((samples.shape[0]*samples.shape[1],samples.shape[2]))
    samples        = rotate_points(samples,angles=angles,offset=center)
    return samples

def batch_of_rays(entry,exit):
    """
    Definition to generate a batch of rays with given entry point(s) and exit point(s). Note that the mapping is one to one, meaning nth item in your entry points list will exit from nth item in your exit list and generate that particular ray. Note that you can have a combination like nx3 points for entry or exit and 1 point for entry or exit. But if you have multiple points both for entry and exit, the number of points have to be same both for entry and exit.

    Parameters
    ----------
    entry      : ndarray
                 Either a single point with size of 3 or multiple points with the size of nx3.
    exit       : ndarray
                 Either a single point with size of 3 or multiple points with the size of nx3.

    Returns
    ----------
    rays       : ndarray
                 Generated batch of rays.
    """
    norays = np.array([0,0])
    if len(entry.shape) == 1:
        entry = entry.reshape((1,3))
    if len(exit.shape) == 1:
        exit = exit.reshape((1,3))
    norays = np.amax(np.asarray([entry.shape[0],exit.shape[0]]))
    if norays > exit.shape[0]:
        exit = np.repeat(exit,norays,axis=0)
    elif norays > entry.shape[0]:
        entry = np.repeat(entry,norays,axis=0)
    rays   = []
    norays = int(norays)
    for i in range(norays):
        rays.append(
                    create_ray_from_two_points(
                                               entry[i],
                                               exit[i]
                                              )
                   )
    rays   = np.asarray(rays)
    return rays



from odak import tools
from odak import np
from odak import wave
import sys

def test():
    no0                  = [100,100]
    no1                  = [100,100]
    size0                = [10.,10.]
    size1                = [10.,10.]
    wavelength           = 0.005
    surface_location_0   = [0.,0.,0.]
    surface_location_1   = [0.,0.,10.]
    wave_number          = wave.wavenumber(wavelength)
    samples_surface_0    = tools.grid_sample(no=no0,size=size0,center=surface_location_0)
    samples_surface_1    = tools.grid_sample(no=no1,size=size1,center=surface_location_1)
    field_0              = np.zeros((no0[0],no0[1]))
    field_0[50,50]       = 1.
    field_0[0,20]        = 2.
    field_0              = field_0.reshape((no0[0]*no0[1]))
    # Propagates to a specific plane.
    field_1              = wave.propagate_field(
                                                samples_surface_0,
                                                samples_surface_1,
                                                field_0,
                                                wave_number,
                                                direction=1
                                               )
    # Reconstruction: propagates back from that specific plane to start plane.
    field_2              = wave.propagate_field(
                                                samples_surface_1,
                                                samples_surface_0,
                                                field_1,
                                                wave_number,
                                                direction=-1
                                               )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



import odak.raytracing as raytracer
import sys
import numpy as np
import odak
from odak.visualize import PLY_object

def test():
    cad = odak.visualize.PLY_object()
    for i in np.linspace(-10,10,5):
        for j in np.linspace(-10,10,5):
            color = [
                     np.int(np.random.rand()*255),
                     np.int(np.random.rand()*255),
                     np.int(np.random.rand()*255)
                    ]
            cad.draw_a_ray([0.,0.,0.],[i,j,50.],color=color)
    cad.save_PLY(savefn='out.ply')
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



"""
``odak.jones``
===================
Provides necessary definitions for jones calculus. See "Introduction to Fourier Optics" by Joseph Goodman.
"""
from odak import np

def electricfield(px,py):
    """
    Definition to create an electric field vector (polarization vector).

    Parameters
    ----------
    px           : float
                   Amplitude of the electric field along X axis.
    py           : float
                   Amplitude of the electric field along Y axis.

    Returns
    ----------
    field        : ndarray
                   An electric field vector (polarization vector).
    """
    field = np.array([[px],[py]])
    return field

def linearpolarizer(field,rotation=0):
    """
    Definition that represents a linear polarizer.

    Parameters
    ----------
    field        : ndarray
                   Polarization vector of an input beam.
    rotation     : float
                   Represents rotation of the polarizer along propagation direction in angles (couter-clockwise).

    Returns
    ----------
    result       : ndarray
                   Polarization vector of an output beam.
    """
    rotation        = np.radians(rotation)
    rotmat          = np.array(
                               [
                                [ float(np.cos(rotation)), float(np.sin(rotation))],
                                [float(-np.sin(rotation)), float(np.cos(rotation))]
                               ]
                              )
    linearpolarizer = np.array([[1,0],[0,0]])
    linearpolarizer = np.dot(rotmat.transpose(),np.dot(linearpolarizer,rotmat))
    result          = np.dot(linearpolarizer,field)
    return result



from odak import np
import torch,torch.nn

def quantize(image_field,bits=4):
    """ 
    Definition to quantize a image field (0-255, 8 bit) to a certain bits level.

    Parameters
    ----------
    image_field : torch.tensor
                  Input image field.
    bits        : int
                  A value in between 0 to 8. Can not be zero.

    Returns
    ----------
    new_field   : torch.tensor
                  Quantized image field.
    """
    divider   = 2**(8-bits)
    new_field = image_field/divider
    new_field = new_field.int()
    return new_field

def zero_pad(field,size=None,method='center'):
    """
    Definition to zero pad a MxN array to 2Mx2N array.

    Parameters
    ----------
    field             : ndarray
                        Input field MxN array.
    size              : list
                        Size to be zeropadded.
    method            : str
                        Zeropad either by placing the content to center or to the left.

    Returns
    ----------
    field_zero_padded : ndarray
                        Zeropadded version of the input field.
    """
    if type(size) == type(None):
        hx = int(torch.ceil(torch.tensor([field.shape[0]/2])))
        hy = int(torch.ceil(torch.tensor([field.shape[1]/2])))
    else:
        hx = int(torch.ceil(torch.tensor([(size[0]-field.shape[0])/2])))
        hy = int(torch.ceil(torch.tensor([(size[1]-field.shape[1])/2])))
    if method == 'center':
        m = torch.nn.ZeroPad2d((hy, hy, hx, hx))
    elif method == 'left aligned':
        m = torch.nn.ZeroPad2d((0, hy*2, 0, hx*2))
    field_zero_padded = m(field)
    if type(size) != type(None):
        field_zero_padded = field_zero_padded[0:size[0],0:size[1]]
    return field_zero_padded

def crop_center(field,size=None):
    """
    Definition to crop the center of a field with 2Mx2N size. The outcome is a MxN array.

    Parameters
    ----------
    field       : ndarray
                  Input field 2Mx2N array.

    Returns
    ----------
    cropped     : ndarray
                  Cropped version of the input field.
    """
    if type(size) == type(None):
        qx      = int(torch.ceil(torch.tensor(field.shape[0])/4))
        qy      = int(torch.ceil(torch.tensor(field.shape[1])/4))
        cropped = field[qx:3*qx,qy:3*qy]
    else:
        cx      = int(torch.ceil(torch.tensor(field.shape[0]/2)))
        cy      = int(torch.ceil(torch.tensor(field.shape[1]/2)))
        hx      = int(torch.ceil(torch.tensor(size[0]/2)))
        hy      = int(torch.ceil(torch.tensor(size[1]/2)))
        cropped = field[cx-hx:cx+hx,cy-hy:cy+hy]
    return cropped

def convolve2d(field,kernel):
    """
    Definition to convolve a field with a kernel by multiplying in frequency space.

    Parameters
    ----------
    field       : torch.tensor
                  Input field with MxN shape.
    kernel      : torch.tensor
                  Input kernel with MxN shape.

    Returns
    ----------
    new_field   : torch.tensor
                  Convolved field.
    """
    fr        = torch.fft.fft2(field)
    fr2       = torch.fft.fft2(torch.flip(torch.flip(kernel,[1,0]),[0,1]))
    m,n       = fr.shape
    new_field = torch.real(torch.fft.ifft2(fr*fr2))
    new_field = torch.roll(new_field,shifts=(int(n/2+1),0),dims=(1,0))
    new_field = torch.roll(new_field,shifts=(int(m/2+1),0),dims=(0,1))
    return new_field

def generate_2d_gaussian(kernel_length=[21,21], nsigma=[3,3]):
    """
    Generate 2D Gaussian kernel. Inspired from https://stackoverflow.com/questions/29731726/how-to-calculate-a-gaussian-kernel-matrix-efficiently-in-numpy

    Parameters
    ----------
    kernel_length : list
                    Length of the Gaussian kernel along X and Y axes.
    nsigma        : list
                    Sigma of the Gaussian kernel along X and Y axes.

    Returns
    ----------
    kernel_2d     : torch.tensor
                    Generated Gaussian kernel.
    """
    x           = torch.linspace(-nsigma[0], nsigma[0], kernel_length[0]+1)
    y           = torch.linspace(-nsigma[1], nsigma[1], kernel_length[1]+1)
    xx, yy      = torch.meshgrid(x, y)
    nsigma      = torch.tensor(nsigma)
    kernel_2d   = torch.exp(-0.5*(torch.square(xx)/torch.square(nsigma[0]) + torch.square(yy)/torch.square(nsigma[1])))
    kernel_2d   = kernel_2d/kernel_2d.sum()
    return kernel_2d

def blur_gaussian(field,kernel_length=[21,21],nsigma=[3,3]):
    """
    A definition to blur a field using a Gaussian kernel.

    Parameters
    ----------
    field         : ndarray
                    MxN field.
    kernel_length : list
                    Length of the Gaussian kernel along X and Y axes.
    nsigma        : list
                    Sigma of the Gaussian kernel along X and Y axes.

    Returns
    ----------
    blurred_field : ndarray
                    Blurred field.
    """
    kernel        = generate_2d_gaussian(kernel_length,nsigma).to(field.device)
    kernel        = zero_pad(kernel,field.shape)
    blurred_field = convolve2d(field,kernel)
    blurred_field = blurred_field/torch.amax(blurred_field)
    return blurred_field



from odak import np
from odak.raytracing import create_ray_from_angles
from odak.tools.transformation import rotate_point,rotate_points
from odak.tools.vector import same_side,point_to_ray_distance

def define_plane(point,angles=[0.,0.,0.]):
    """ 
    Definition to generate a rotation matrix along X axis.

    Parameters
    ----------
    point        : ndarray
                   A point that is at the center of a plane.
    angles       : list
                   Rotation angles in degrees.

    Returns
    ----------
    plane        : ndarray
                   Points defining plane.
    """
    plane = np.array([
                      [ 10., 10., 0.],
                      [  0., 10., 0.],
                      [  0.,  0., 0.]
                     ],dtype=np.float)
    point = np.asarray(point)
    for i in range(0,plane.shape[0]):
        plane[i],_,_,_  = rotate_point(plane[i],angles=angles)
        plane[i]        = plane[i]+point
    return plane

def bring_plane_to_origin(point,plane,shape=[10.,10.],center=[0.,0.,0.],angles=[0.,0.,0.],mode='XYZ'):
    """
    Definition to bring points back to reference origin with respect to a plane.

    Parameters
    ----------
    point              : ndarray
                         Point(s) to be tested.
    shape              : list
                         Dimensions of the rectangle along X and Y axes.
    center             : list
                         Center of the rectangle.
    angles             : list
                         Rotation angle of the rectangle.
    mode               : str
                         Rotation mode of the rectangle, for more see odak.tools.rotate_point and odak.tools.rotate_points.

    Returns
    ----------
    transformed_points : ndarray
                         Point(s) that are brought back to reference origin with respect to given plane.
    """
    if point.shape[0] == 3:
        point = point.reshape((1,3))
    reverse_mode       = mode[::-1]
    angles             = [-angles[0],-angles[1],-angles[2]]
    center             = np.asarray(center).reshape((1,3))
    transformed_points = point-center
    transformed_points = rotate_points(
                                       transformed_points,
                                       angles=angles,
                                       mode=reverse_mode,
                                      )
    if transformed_points.shape[0] == 1:
        transformed_points = transformed.points.reshape((3,))
    return transformed_points


def center_of_triangle(triangle):
    """
    Definition to calculate center of a triangle.

    Parameters
    ----------
    triangle      : ndarray
                    An array that contains three points defining a triangle (Mx3). It can also parallel process many triangles (NxMx3).
    """
    if len(triangle.shape) == 2:
        triangle = triangle.reshape((1,3,3))
    center = np.mean(triangle,axis=1)
    return center

def is_it_on_triangle(pointtocheck,point0,point1,point2):
    """
    Definition to check if a given point is inside a triangle. If the given point is inside a defined triangle, this definition returns True.

    Parameters
    ----------
    pointtocheck  : list
                    Point to check.
    point0        : list
                    First point of a triangle.
    point1        : list
                    Second point of a triangle.
    point2        : list
                    Third point of a triangle.
    """
    # point0, point1 and point2 are the corners of the triangle.
    pointtocheck = np.asarray(pointtocheck).reshape(3)
    point0 = np.asarray(point0); point1 = np.asarray(point1); point2 = np.asarray(point2)
    side0        = same_side(pointtocheck,point0,point1,point2)
    side1        = same_side(pointtocheck,point1,point0,point2)
    side2        = same_side(pointtocheck,point2,point0,point1)
    if side0 == True and side1 == True and side2 == True:
        return True
    return False

def define_circle(center,radius,angles):
    """
    Definition to describe a circle in a single variable packed form.

    Parameters
    ----------
    center  : float
              Center of a circle to be defined.
    radius  : float
              Radius of a circle to be defined.
    angles  : float
              Angular tilt of a circle.

    Returns
    ----------
    circle  : list
              Single variable packed form.
    """
    points  = define_plane(center,angles=angles)
    circle  = [
               points,
               center,
               radius
              ]
    return circle

def define_sphere(center,radius):
    """
    Definition to define a sphere.

    Parameters
    ----------
    center     : ndarray
                 Center of a sphere in X,Y,Z.
    radius     : float
                 Radius of a sphere.

    Returns
    ----------
    sphere     : ndarray
                 Single variable packed form.
    """
    sphere = np.array([center[0],center[1],center[2],radius],dtype=np.float)
    return sphere

def sphere_function(point,sphere):
    """
    Definition of a sphere function. Evaluate a point against a sphere function.

    Parameters
    ----------
    sphere     : ndarray
                 Sphere parameters, XYZ center and radius.
    point      : ndarray
                 Point in XYZ.

    Return
    ----------
    result     : float
                 Result of the evaluation. Zero if point is on sphere.
    """
    point = np.asarray(point)
    if len(point.shape) == 1:
        point = point.reshape((1,3)) 
    result = (point[:,0]-sphere[0])**2 + (point[:,1]-sphere[1])**2 + (point[:,2]-sphere[2])**2 - sphere[3]**2
    return result

def define_cylinder(center,radius,rotation=[0.,0.,0.]):
    """
    Definition to define a cylinder

    Parameters
    ----------
    center     : ndarray
                 Center of a cylinder in X,Y,Z.
    radius     : float
                 Radius of a cylinder along X axis.
    rotation   : list
                 Direction angles in degrees for the orientation of a cylinder.

    Returns
    ----------
    cylinder   : ndarray
                 Single variable packed form.
    """
    cylinder_ray = create_ray_from_angles(np.asarray(center),np.asarray(rotation))
    cylinder     = np.array(
                            [
                             center[0],
                             center[1],
                             center[2],
                             radius,
                             center[0]+cylinder_ray[1,0],
                             center[1]+cylinder_ray[1,1],
                             center[2]+cylinder_ray[1,2]
                            ],
                            dtype=np.float
                            )
    return cylinder

def cylinder_function(point,cylinder):
    """
    Definition of a cylinder function. Evaluate a point against a cylinder function. Inspired from https://mathworld.wolfram.com/Point-LineDistance3-Dimensional.html

    Parameters
    ----------
    sphere     : ndarray
                 Sphere parameters, XYZ center and radius.
    point      : ndarray
                 Point in XYZ.

    Return
    ----------
    result     : float
                 Result of the evaluation. Zero if point is on sphere.
    """
    point = np.asarray(point)
    if len(point.shape) == 1:
        point = point.reshape((1,3))
    distance = point_to_ray_distance(
                                     point,
                                     np.array([cylinder[0],cylinder[1],cylinder[2]],dtype=np.float),
                                     np.array([cylinder[4],cylinder[5],cylinder[6]],dtype=np.float)
                                    )
    r        = cylinder[3]
    result   = distance-r**2
    return result



import odak.raytracing as raytracer
import sys
from odak.tools.asset import read_PLY,write_PLY

def test():
    triangles = read_PLY('./test/sample.ply')
    centers   = raytracer.center_of_triangle(triangles) 
    normals   = raytracer.get_triangle_normal(triangles,centers)
    write_PLY(triangles,savefn='output.ply')
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



#!/usr/bin/env python

import sys
import numpy as np
import odak
import odak.raytracing

def test_ray_intersect_w_a_triangle():
    # Input points to create a ray.
    ray_start_point_0   = [5.,5.,0.]
    ray_end_point_0     = [10.,10.,1000.]
    # Create from two points.
    ray_0               = odak.raytracing.create_ray_from_two_points(
                                                                     ray_start_point_0,
                                                                     ray_end_point_0
                                                                    )
    # Input points to create a ray.
    ray_start_point_1   = [5.,5.,0.]
    ray_end_point_1     = [0.,100.,1000.]
    # Create from two points.
    ray_1               = odak.raytracing.create_ray_from_two_points(
                                                                     ray_start_point_1,
                                                                     ray_end_point_1
                                                                    )
    # Intersection with a triangle.
    triangle            = [
                           [  50.,  50., 1000.],
                           [  -5.,  -5., 1000.],
                           [   0.,  50., 1000.],
                          ]
    normal_0,distance_0 = odak.raytracing.intersect_w_triangle(
                                                               ray_0,
                                                               triangle
                                                              )
    normal_1,distance_1 = odak.raytracing.intersect_w_triangle(
                                                               ray_1,
                                                               triangle
                                                              )
    print(normal_0)
    print(normal_1)
    assert True==True

if __name__ == '__main__':
    sys.exit(test_ray_intersect_w_a_triangle())



import bpy,queue
import threading
import os,socket,traceback,sys
sys.path.append('%s' % os.path.dirname(os.path.realpath(__file__)))
from libblend import *

execution_queue = queue.Queue()

class whisper2blender(threading.Thread):
    def __init__(self,port=8082,host='127.0.0.1',path_max=4096):
        threading.Thread.__init__(self)
        self.port     = port
        self.host     = host
        self.path_max = path_max
    def run(self):
        serversocket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        serversocket.bind((self.host, self.port))
        serversocket.listen(1)
        print("Listening on %s:%s" % (self.host, self.port))
        while True:
            connection, address = serversocket.accept()
            buf                 = connection.recv(self.path_max)
            for cmd in buf.split(b'\x00'):
                if cmd:
                    try:
                        cmd = cmd.decode()
#                        print(cmd)
                        run_in_main_thread(cmd)
    #                    exec(cmd)
                    except:
                        traceback.print_exc()

def run_in_main_thread(function):
    execution_queue.put(function)

def execute_queued_functions():
    while not execution_queue.empty():
        function = execution_queue.get()
        if type(function) != type(None):
            exec(function)
    return 0.1

def main():
    messenger = whisper2blender()
    messenger.start()
    bpy.app.timers.register(execute_queued_functions)
    return True

if __name__ == "__main__":
    main()




"""
``odak.tools``
===================
Provides necessary definitions for general tools used across the library.

"""
# To get sub-modules.
from .transformation import *
from .vector import *
from .sample import *
from .asset import *
from .file import *
from .matrix import *
from .conversions import *



from odak import np

def line_spread_function(line):
    """
    Definition to take the gradient of a 1D function.

    Parameters
    ----------
    line          : ndarray
                    1D array.

    Returns
    ----------
    result        : ndarray
                    Gradient of the given 1D array.
    """
    if np.__name__ == 'cupy':
       import numpy
       cache   = np.asnumpy(line)
       result  = numpy.gradient(cache)
    else:
       result  = np.gradient(line)
    result  = np.asarray(result)
    return result

def fourier_transform_1d(line):
    """
    Definition to take the 1D fourier transform. This is used only for modulation transfer function calculations.

    Parameters
    ----------
    line         : ndarray
                   1D array.

    Returns
    ----------
    result       : ndarray
                   Positive side of the fourier transform of the given line.
    """
    result  = np.fft.fft(line) #/len(der_x)
    result /= np.amax(result)
    result  = result[np.arange(0,int(line.shape[0]/2))]
    return result

def polynomial_fit(line_x,line_y,fit_degree=3):
    """
    Definition to fit polynomials to a vector.

    Parameters
    ----------
    line_x      : ndarray
                  Values along X axis.
    line_y      : ndarray
                  Values along Y axis.
    degree      : int
                  Degree of the polynomial fit.
    
    Returns
    ----------
    p           : numpy.poly1d
                  polynomial fit.
    """
    if np.__name__ == 'numpy':
        fun_poly = np.polyfit(line_x,line_y,fit_degree)
        p        = np.poly1d(fun_poly)
    else:
        import numpy
        line_x   = np.asnumpy(line_x)
        line_y   = np.asnumpy(line_y)
        fun_poly = numpy.polyfit(line_x,line_y,fit_degree)
        p        = numpy.poly1d(fun_poly)
    return p

def roi(image,location=[0,100,0,100],threshold=[0,1,0,1]):
    """
    Definition to get the lines from a target ROI.

    Parameters
    ----------
    image      : ndarray
                 a 2D image to be sliced (nxm).
    location   : ndarray
                 Locations for taking the ROI.
    threshold  : list
                 Threshold below and above these numbers.

    Returns
    -------
    line_x     : ndarray
                 Line slice.
    line_y     : ndarray
                 Line slice.
    """
    img                         = image[location[0]:location[1],location[2]:location[3]]
    if len(img.shape) == 3:
        img = np.sum(img,axis=2)
    line_x                      = img[:,int(img.shape[1]/2)]
    line_y                      = img[int(img.shape[0]/2),:]
    line_x                      = np.asarray(line_x)
    line_y                      = np.asarray(line_y)
    line_x                      = line_x-np.amin(line_x)
    line_x                      = line_x/np.amax(line_x)
    line_y                      = line_y-np.amin(line_y)
    line_y                      = line_y/np.amax(line_y)    
    line_x[line_x<threshold[0]] = 0
    line_x[line_x>threshold[1]] = 1
    line_y[line_y<threshold[2]] = 0
    line_y[line_y>threshold[3]] = 1
    return line_x,line_y,img

def modulation_transfer_function(line_x,line_y,px_size):
    """
    Definition to compute modulation transfer function. This definition is based on the work by Peter Burns. For more consult Burns, Peter D. "Slanted-edge MTF for digital camera and scanner analysis." Is and Ts Pics Conference. SOCIETY FOR IMAGING SCIENCE & TECHNOLOGY, 2000.

    Parameters
    ----------
    line_x       : ndarray
                   Slice from an image along X axis. See odak.measurements.roi().
    line_y       : ndarray
                   Slice from an image along Y axis. See odak.measurements.roi().
    px_size      : ndarray
                   Physical angular sizes that each pixels corresponds to on the image plane both for X and Y axes.

    Returns
    ----------
    mtf          : ndarray
                   Calculated modulation transfer function along X and Y axes.
    frq          : ndarray
                   Frequencies of the calculated MTF.
    """
    # 1st derivative of both values: "Line Spread Function",
    der_x         = line_spread_function(line_x)
    der_y         = line_spread_function(line_y)
    # Fourier transform of the first derivative: "Modulation Transfer Function",
    mtf_x         = fourier_transform_1d(der_x)
    mtf_y         = fourier_transform_1d(der_y)
    # Arrange the corresponding frequencies,
    n_x           = len(der_x) # length of the signal,
    k_x           = np.arange(n_x)
    T_x           = n_x*px_size[0]
    frq_x         = k_x/T_x
    frq_x         = frq_x[np.arange(0,int(n_x/2))]
    n_y           = len(der_y) # length of the signal,
    k_y           = np.arange(n_y)
    T_y           = n_y*px_size[1]
    frq_y         = k_y/T_y
    frq_y         = frq_y[np.arange(0,int(n_y/2))]
    return [mtf_x,mtf_y],[frq_x,frq_y]



from odak import np
import odak.catalog
from odak.raytracing.primitives import define_plane,bring_plane_to_origin
from odak.raytracing.boundary import intersect_w_surface
from odak.wave import wavenumber
from odak.tools.sample import circular_uniform_sample,grid_sample
from odak.tools.transformation import rotate_points,tilt_towards
from odak.raytracing.ray import create_ray_from_two_points

class thin_diffuser():
    """
    A class to represent a thin diffuser. This is generally useful for raytracing and wave calculations.
    """
    def __init__(self,phase=None,shape=[10.,10.],center=[0.,0.,0.],angles=[0.,0.,0.],diffusion_angle=5.,diffusion_no=[3,3],name='diffuser'):
        """
        Class to represent a simple thin diffuser with a random phase.

        Parameters
        ----------
        phase            : ndarray
                           Initial phase to be loaded. If non provided, it will start with a random phase.
        shape            : list
                           Shape of the detector.
        center           : list
                           Center of the detector.
        angles           : list
                           Rotation angles of the detector in degrees.
        diffusion angles : list
                           Full angle of diffusion along two axes.
        diffusion_no     : list
                           Number of rays to be generated along two axes at each diffusion.
        """
        self.settings           = {
                                   'name'                     : name,
                                   'center'                   : center,
                                   'angles'                   : angles,
                                   'rotation mode'            : 'XYZ',
                                   'shape'                    : shape,
                                   'diffusion angle'          : diffusion_angle,
                                   'number of diffusion rays' : diffusion_no
                                  }
        self.plane              = define_plane(
                                               self.settings['center'],
                                               angles=self.settings['angles']
                                              )
        self.k                  = wavenumber(0.05)
        self.diffusion_points   = circular_uniform_sample(
                                                          no=self.settings["number of diffusion rays"],
                                                          radius=np.tan(np.radians(self.settings["diffusion angle"]/2.)),
                                                          center=[0.,0.,1.]
                                                         )
        if type(phase) == type(None):
           self.surface_points = grid_sample(
                                             no=[100,100],
                                             size=self.settings["shape"],
                                             center=self.settings["center"],
                                             angles=self.settings["angles"]
                                            )
    def plot_diffuser(self,figure):
        """
        Definition to plot diffuser to a odak.visualize.plotly.rayshow().

        Parameters
        ----------
        figure      : odak.visualize.plotly.rayshow()
                      Figure to plot the diffuser.
        """
        points = grid_sample(
                             no=[10,10],
                             size=self.settings["shape"],
                             center=self.settings["center"],
                             angles=self.settings["angles"]
                            )
        points = points.reshape((10,10,3))
        figure.add_surface(
                           data_x=points[:,:,0],
                           data_y=points[:,:,1],
                           data_z=points[:,:,2],
                           surface_color=np.zeros(points[:,:,2].shape),
                           opacity=0.5,
                           contour=False
                          )

    def raytrace(self,ray):
        """
        Definition to raytrace the diffuser.


        Parameters
        ----------
        ray         : ndarray
                      Ray(s).

        Returns
        ----------
        new_rays    : ndarray
                      Diffusion rays.
        normal      : ndarray
                      Surface normals
        distance    : ndarray 
                      Distance ray(s) has/have travelled until to the point of diffusion.
        """
        if len(ray.shape) == 2:
            ray = ray.reshape((1,ray.shape[0],ray.shape[1]))
        normal,distance        = intersect_w_surface(ray,self.plane)
        ########################################################################
        # This is the bottleneck for has to be replaced for better performance #
        ########################################################################
        new_rays               = np.empty((0,2,3))
        for point_id in range(0,normal.shape[0]):
            tilt_angles   = tilt_towards(ray[point_id,0],normal[point_id,0])  
            tilted_points = rotate_points(
                                          self.diffusion_points,
                                          angles=tilt_angles,
                                          offset=normal[point_id,0]
                                         )
            new_rays      = np.vstack((new_rays,create_ray_from_two_points(normal[point_id,0],tilted_points)))
        new_rays = np.asarray(new_rays)
        return new_rays,normal,distance



import sys

def test():
    import torch
    from odak.learn.wave import stochastic_gradient_descent
    wavelength               = 0.000000532
    dx                       = 0.0000064
    distance                 = 0.1
    resolution               = [1080,1920]
    cuda                     = False
    device                   = torch.device("cuda" if cuda else "cpu")
    input_field              = torch.zeros(resolution[0],resolution[1]).to(device)
    input_field[500::600,:]  = 1
    iteration_number         = 50
    hologram,reconstructed   = stochastic_gradient_descent(
                                                           input_field,
                                                           wavelength,
                                                           distance,
                                                           dx,
                                                           resolution,
                                                           'TR Fresnel',
                                                           iteration_number,
                                                           learning_rate=0.1,
                                                           cuda=cuda
                                                          )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



from odak import np

def double_convergence(nx,ny,k,r,dx):
    """
    A definition to generate initial phase for a Gerchberg-Saxton method. For more details consult Sun, Peng, et al. "Holographic near-eye display system based on double-convergence light Gerchberg-Saxton algorithm." Optics express 26.8 (2018): 10140-10151.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    r          : float
                 The distance between location of a light source and an image plane.
    dx         : float
                 Pixel pitch.

    Returns
    ---------
    function   : ndarray
                 Generated phase pattern for a Gerchberg-Saxton method.
    """
    size = [ny,nx]
    x    = np.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])
    y    = np.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])
    X,Y  = np.meshgrid(x,y)
    Z    = X**2+Y**2  
    w    = np.exp(1j*k*Z/r)
    return w

def quadratic_phase_function(nx,ny,k,focal=0.4,dx=0.001,offset=[0,0]):
    """ 
    A definition to generate 2D quadratic phase function, which is typically use to represent lenses.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    focal      : float
                 Focal length of the quadratic phase function.
    dx         : float
                 Pixel pitch.
    offset     : list
                 Deviation from the center along X and Y axes.

    Returns
    ---------
    function   : ndarray
                 Generated quadratic phase function.
    """
    size = [ny,nx]
    x    = np.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])-offset[1]*dx
    y    = np.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])-offset[0]*dx
    X,Y  = np.meshgrid(x,y)
    Z    = X**2+Y**2
    qwf  = np.exp(1j*k*0.5*np.sin(Z/focal))
    return qwf

def prism_phase_function(nx,ny,k,angle,dx=0.001,axis='x'):
    """
    A definition to generate 2D phase function that represents a prism. See Goodman's Introduction to Fourier Optics book for more.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    angle      : float
                 Tilt angle of the prism in degrees.
    dx         : float
                 Pixel pitch.
    axis       : str
                 Axis of the prism.

    Returns
    ----------
    prism      : ndarray
                 Generated phase function for a prism.
    """
    angle = np.radians(angle)
    size  = [ny,nx]
    x     = np.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])
    y     = np.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])
    X,Y   = np.meshgrid(x,y)
    if axis == 'y':
        prism = np.exp(-1j*k*np.sin(angle)*Y)
    elif axis == 'x':
        prism = np.exp(-1j*k*np.sin(angle)*X)
    return prism

def freeform(nx,ny,k,distances,dx=0.001):
    """
    A definition to generate a freeform field pattern from a depth map.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    distances  : ndarray
                 Depth map.
    dx         : float
                 Pixel pitch.

    Returns
    ---------
    field      : ndarray
                 Generated pattern.
    """
    size  = [ny,nx]
    x     = np.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])
    y     = np.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])
    X,Y   = np.meshgrid(x,y)
    Z     = X**2+Y**2
    field = np.exp(1j*k*0.5*np.sin(Z/distances))
    return field

def plane_tilt(nx,ny,k,focals,dx=0.001,axis='x'):
    """
    A definition to tilt a complex field.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    k          : odak.wave.wavenumber
                 See odak.wave.wavenumber for more.
    focals     : list
                 Focus ranges, two for X and two for Y, in total four numbers. Make sure to pass nonzero numbers.
    dx         : float
                 Pixel pitch.
    axis       : str
                 Which state to tilt, x, y or xy.

    Returns
    ----------
    field      : ndarray
                 Field to tilt a plane.
    """
    size       = [ny,nx]
    x          = np.linspace(-size[0]*dx/2,size[0]*dx/2,size[0])
    y          = np.linspace(-size[1]*dx/2,size[1]*dx/2,size[1])
    X,Y        = np.meshgrid(x,y)
    Z          = X**2+Y**2
    focals     = np.asarray(focals)
    if np.where(focals==0)[0].shape[0] != 0:
        raise Exception("Focals must be non zero.")
    if np.__name__ == 'cupy':
        import numpy as np_cpu
        focals = np.asnumpy(focals)
    else:
        np_cpu = np
    focal_x    = np_cpu.geomspace(focals[0],focals[1],size[0])
    focal_y    = np_cpu.geomspace(focals[2],focals[3],size[1])
    if np.__name__ == 'cupy':
        focal_x = np.asarray(focal_x)
        focal_y = np.asarray(focal_y)
    FX,FY      = np.meshgrid(focal_x,focal_y)
    field      = np.ones((nx,ny),dtype=np.complex64)
    if axis == 'x' or axis == 'xy':
        field *= np.exp(1j*k*0.5*np.sin(Z/FX))
    if axis == 'y' or axis == 'xy':
        field *= np.exp(1j*k*0.5*np.sin(Z/FY))
    return field

def linear_grating(nx,ny,every=2,add=3.14,axis='x'):
    """
    A definition to generate a linear grating.

    Parameters
    ----------
    nx         : int
                 Size of the output along X.
    ny         : int
                 Size of the output along Y.
    every      : int
                 Add the add value at every given number.
    add        : float
                 Angle to be added.
    axis       : string
                 Axis eiter X,Y or both.

    Returns
    ----------
    field      : ndarray
                 Linear grating term.
    """
    grating = np.zeros((nx,ny),dtype=np.complex64)
    if axis == 'x':
        grating[::every,:] = np.exp(1j*add)
    if axis == 'y':
        grating[:,::every] = np.exp(1j*add)
    if axis == 'xy':
        checker  = np.indices((nx,ny)).sum(axis=0) % every
        checker += 1
        checker  = checker % 2
        grating  = np.exp(1j*checker*add)
    return grating



import sys
import os
import odak
from odak import np
from odak.wave import wavenumber,propagate_beam
from odak.learn.wave import propagate_beam as propagate_beam_torch
from odak.tools import zero_pad
import torch

def compare():
    wavelength                 = 0.5*pow(10,-6)
    pixeltom                   = 6*pow(10,-6)
    distance                   = 0.2
    propagation_type           = 'TR Fresnel'
    k                          = wavenumber(wavelength)
    sample_field               = np.zeros((500,500),dtype=np.complex64)
    sample_field[
                 240:260,
                 240:260
                ]              = 1000
    random_phase               = np.pi*np.random.random(sample_field.shape)
    sample_field               = sample_field*np.cos(random_phase)+1j*sample_field*np.sin(random_phase)
    sample_field               = zero_pad(sample_field)

    if np.__name__ == 'cupy':
        sample_field = np.asnumpy(sample_field)

    sample_field_torch         = torch.from_numpy(sample_field)

    ## Propagate and reconstruct using torch.
    hologram_torch             = propagate_beam_torch(
                                                      sample_field_torch,
                                                      k,
                                                      distance,
                                                      pixeltom,
                                                      wavelength,
                                                      propagation_type
                                                     )
    reconstruction_torch       = propagate_beam_torch(
                                                      hologram_torch,
                                                      k,
                                                      -distance,
                                                      pixeltom,
                                                      wavelength,
                                                      propagation_type
                                                     )   

    ## Propagate and reconstruct using np.
    hologram                   = propagate_beam(
                                                sample_field,
                                                k,
                                                distance,
                                                pixeltom,
                                                wavelength,
                                                propagation_type
                                               )
    reconstruction             = propagate_beam(
                                                hologram,
                                                k,
                                                -distance,
                                                pixeltom,
                                                wavelength,
                                                propagation_type
                                               )
    np.testing.assert_array_almost_equal(hologram_torch.numpy(), hologram, 5)
    np.testing.assert_array_almost_equal(reconstruction_torch.numpy(), reconstruction, 5)
    assert True==True

if __name__ == '__main__':
    sys.exit(compare())



#!/usr/bin/env python

import sys

def test_cylinder_intersection():
    import odak.raytracing as raytracer
    ray             = raytracer.create_ray_from_two_points([0.,0.,0.],[10.,0.,0.])
    cylinder        = raytracer.define_cylinder([0.,0.,0.],5.,rotation=[0.,90.,0.])
    normal,distance = raytracer.intersect_w_cylinder(ray,cylinder)
    assert True==True

def test_multiple_rays_w_cylinder():
    import odak.raytracing as raytracer
    import odak.tools as tools
    end_points        = tools.grid_sample(
                                          no=[5,5],
                                          size=[10.,10.],
                                          center=[0.,0.,100.],
                                          angles=[0.,0.,0.]
                                         )
    start_point       = [0.,0.,0.]
    rays              = raytracer.create_ray_from_two_points(
                                                             start_point,
                                                             end_points
                                                            )
    sphere            = raytracer.define_sphere([0.,0.,100.],20)
    normals,distances = raytracer.intersect_w_sphere(rays,sphere)
    assert True==True

def test_all():
    test_cylinder_intersection()
#    test_multiple_rays_w_sphere()

if __name__ == '__main__':
    sys.exit(test_all())



import odak.catalog
from odak.wave import calculate_phase,calculate_amplitude,calculate_intensity
from odak.raytracing.primitives import define_plane,bring_plane_to_origin
from odak.raytracing.boundary import intersect_w_surface
from odak.visualize.plotly import detectorshow
from odak.tools.sample import grid_sample
from odak import np

class plane_detector():
    """
    A class to represent a plane detector. This is generally useful in raytracing and wave calculations.
    """
    def __init__(self,field=None,resolution=[1000,1000],shape=[10.,10.],center=[0.,0.,0.],angles=[0.,0.,0.],name='detector'):
        """
        Class to represent a simple planar detector.

        Parameters
        ----------
        field       : ndarray
                      Initial field to be loaded.
        resolution  : list
                      Resolution of the detector.
        shape       : list
                      Shape of the detector.
        center      : list
                      Center of the detector.
        angles      : list
                      Rotation angles of the detector.
        """
        self.settings   = {
                           'name'          : name,
                           'resolution'    : resolution,
                           'center'        : center,
                           'angles'        : angles,
                           'rotation mode' : 'XYZ',
                           'shape'         : shape,
                          }
        self.plane      = define_plane(
                                       self.settings['center'],
                                       angles=self.settings['angles']
                                      )
        if type(field) == type(None):
            self.field = np.zeros(
                                  (
                                   self.settings['resolution'][0],
                                   self.settings['resolution'][1],
                                   1
                                  ),
                                  dtype=np.complex64
                                  )
            self.clear_detector()

    def clear_detector(self):
        """
        A definition to clear the field accumulated on the detector.
        """
        self.field = np.zeros(self.field.shape,dtype=np.complex64)

    def plot_detector(self,figure,channel=0,plot_type='intensity'):
        """ 
        Definition to plot detector to a odak.visualize.plotly.rayshow().

        Parameters
        ----------
        figure      : odak.visualize.plotly.rayshow()
                      Figure to plot the diffuser.
        channel     : int
                      A channel of the current field. 
        plot_type   : str
                      Plot type, it can be either phase, amplitude or intensity.
        """
        points    = grid_sample(
                                no=[self.settings['resolution'][0],self.settings['resolution'][1]],
                                size=self.settings["shape"],
                                center=self.settings["center"],
                                angles=self.settings["angles"]
                               )
        points    = points.reshape(
                                   (
                                    self.settings['resolution'][0],
                                    self.settings['resolution'][1],
                                    3
                                   )
                                  )
        if plot_type == 'phase':
           values = self.get_phase()
        elif plot_type == 'amplitude':
           values = self.get_amplitude()
        elif plot_type == 'intensity':
           values = self.get_intensity()
        figure.add_surface(
                           data_x=points[:,:,0],
                           data_y=points[:,:,1],
                           data_z=points[:,:,2],
                           surface_color=values[:,:,channel],
                           opacity=0.5,
                           contour=False
                          )

    def plot_field(self,channel=0):
        """
        A definition to plot the field measured on the detector using plotly.
        """
        self.fig = detectorshow()
        self.fig.add_field(self.field[:,:,channel])
        self.fig.show()

    def get_field(self):
        """
        A definition to return the field measured on the detector.

        Returns
        ----------
        field       : ndarray
                      A copy of the field measured by the detector.
        """
        return np.copy(self.field)

    def get_intensity(self):
        """
        A definition to return the intensity of the field.

        Returns
        ---------
        intensity   : ndarray
                      Intensity of the field measured by the detector.
        """
        intensity = calculate_intensity(self.field)
        return intensity

    def get_amplitude(self):
        """
        A definition to return the amplitude of the field.

        Returns
        ---------
        amplitude   : ndarray
                      Amplitude of the field measured by the detector.
        """
        amplitude = calculate_amplitude(self.field)
        return amplitude

    def get_phase(self):
        """
        A definition to return the phase of the field.

        Returns
        ---------
        phase       : ndarray
                      Phase of the field measured by the detector.
        """
        phase = calculate_phase(self.field)
        return phase

    def raytrace(self,ray,field=None,channel=0):
        """
        A definition to calculate the intersection between given ray(s) and the detector. If a ray contributes to the detector, field will be taken into account in calculating the field over the planar detector.
 
        Parameters
        ----------
        ray          : ndarray
                       Ray(s) to be intersected.
        field        : ndarray
                       Field(s) to be used for calculating contribution of rays to the detector.
        channel      : list
                       Which color channel to contribute to in the detector plane. Default is zero. One can use a list to select multiple channels separately.
 
        Returns
        ----------
        normal       : ndarray
                       Normal for each intersection point.
        distance     : ndarray
                       Distance for each ray.
        """
        normal,distance = intersect_w_surface(ray,self.plane)
        points          = bring_plane_to_origin(
                                                normal[:,0],
                                                self.plane,
                                                shape=self.settings["shape"],
                                                center=self.settings["center"],
                                                angles=self.settings["angles"],
                                                mode=self.settings["rotation mode"]
                                               )
        if points.shape[0] == 3:
            points = points.reshape((1,3))
        # This could improve with a bilinear filter. Basically removing int with a filter.
        detector_ids = np.array(
                                [
                                 (points[:,0]+self.settings["shape"][0]/2.)/self.settings["shape"][0]*self.settings["resolution"][0]+1,
                                 (points[:,1]+self.settings["shape"][1]/2.)/self.settings["shape"][1]*self.settings["resolution"][1]+1
                                ],
                                dtype=int
                               )
        detector_ids[0,:] = (detector_ids[0,:]>=1)*detector_ids[0,:]
        detector_ids[1,:] = (detector_ids[1,:]>=1)*detector_ids[1,:]
        detector_ids[0,:] = (detector_ids[0,:]<self.settings["resolution"][0]+1)*detector_ids[0,:]
        detector_ids[1,:] = (detector_ids[1,:]<self.settings["resolution"][1]+1)*detector_ids[1,:]
        cache             = np.zeros(
                                     (
                                      self.settings["resolution"][0]+1,
                                      self.settings["resolution"][1]+1,
                                      self.field.shape[2]
                                     ),
                                     dtype=np.complex64
                                    )

        ##################################################################
        # This solution is far from ideal. There has to be a better way. #
        ##################################################################
        for detector_id in range(0,detector_ids.shape[1]):
            x           = detector_ids[0,detector_id]
            y           = detector_ids[1,detector_id]
            r2          = distance[detector_id]**2
            if type(field) == type(None):
                cache[x,y] += 1000./r2
            else:
                cache[x,y] += field[x,y]/r2
        ##################################################################
        # This solution isn't working at all as two same ids are         #
        # interpretted as one.                                           #
        ##################################################################
        #cache[
        #       detector_ids[0],
        #       detector_ids[1],
        #       channel
        #      ]           += field
        ##################################################################
        self.field      += cache[1::,1::,:]
        return normal,distance



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Whole library can be found under https://github.com/kunguz/odak,

import sys,scipy
import matplotlib
#matplotlib.use('Agg')
import matplotlib.pyplot
#import mpl_toolkits.mplot3d.art3d as art3d
import scipy.linalg
from datetime import datetime
#from matplotlib.mlab import griddata
from matplotlib import cm
from matplotlib.patches import Circle, PathPatch, Ellipse
from mpl_toolkits.mplot3d import axes3d
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from numpy import *
from numpy.fft import *
from math import radians,tan,pi

__author__  = ('Kaan Akşit')
debug = True


# Definition to prompt.
def prompt(txt,who='odak',debug=True):
    if debug == True:
        print('[%s] [%s] %s' % (str(datetime.now()),who,txt))

class ParaxialMatrix():
    def __init__(self):
        # See "Laser beams and resonators" from Kogelnik and Li for the theoratical explanation
        self.plt = matplotlib.pyplot
        self.fig = self.plt.figure()
        self.ax  = self.fig.add_subplot(111,aspect='equal')
        # Show grid.
        self.plt.grid()
        return
    def CreateVector(self,x,angle):
        # Creates a paraxial ray, angle is in degrees, x is the distance of the point to the plane of direction of propagation
        angle  = radians(angle)
        vector = array([[x],[tan(angle)],[1]])
        return vector
    def FreeSpace(self,vector,distance,deltax=0,deltafi=0):
        # Ray transfer matrix of free space propagation, distance is in milimeters.
        # deltax is the spatial shift, deltafi is the angular shift.
        space  = array([[1,distance,deltax],[0,1,deltafi],[0,0,1]])
        vector = dot(space,vector)
        return vector
    def CurvedInterface(self,vector,n1,n2,R,deltax=0,deltafi=0):
        # Ray transfer matrix of a curved interface, focal length is f and is in in milimeters.
        # Taken from Wikipedia article: Ray transfer matrix anaylsis.
        # deltax is the spatial shift, deltafi is the angular shift.
        # n1 is the first medium that the ray is coming from.
        # n2 is the second medium that the ray is entering to.
        # R is the radius of curvature, R>0 for convex
        CInter = array([[1,0,deltax],[(n1-n2)/R/n2,n1/n2,deltafi],[0,0,1]])
        vector = dot(CInter,vector)
        return vector
    def PlotVector(self,startvector,stopvector,posx=0,distance=0,color='g+-'):
        if stopvector[1] !=0:
            # Method to plot paraxial vectors in 2D space.
            self.plt.plot([posx,(stopvector[0]-startvector[0])/stopvector[1]+posx],[startvector[0],stopvector[0]],color)
            # Return new position at X-axis.
            posx += (stopvector[0]-startvector[0])/stopvector[1]
        else:
            self.plt.plot([posx,posx+distance],[startvector[0],stopvector[0]],color)
            posx = distance
        return posx
    def PlotLine(self,point1,point2,color='ro--'):
        # Definition to plot a line in between two points.
        self.plt.plot(point1,point2,color)
        return True
    def PlotLens(self,CenterXY, thickness, LensHeight, rotation, alpha=0.5):
        # Definition to plot a lens.
        lens = Ellipse(xy=CenterXY, width=thickness, height=LensHeight, angle=-rotation)
        self.ax.add_artist(lens)
        lens.set_clip_box(self.ax.bbox)
        lens.set_alpha(alpha)
        return True
    def InitNewPlot(self):
        # New plot initiated.
        NewFig  = self.plt.figure()
        NewPlot = NewFig.add_subplot(111)
        return NewPlot,NewFig
    def PlotHist(self,dataset,plot):
        # Definition to plot a histogram.
        plot.hist(dataset,bins=1000,color='blue',normed='True')
        return True
    def PlotData(self,dataset,color,alpha=1,linestyle='-'):
        # Definition to plot a dataset.
        self.plt.plot(dataset[0],dataset[1],linestyle=linestyle,color=color,alpha=alpha)
        return True
    def PlotFillData(self,dataset,color):
        # Definition to plot a dataset with fill.
        self.plt.fill(dataset[0],dataset[1],color,alpha=0.3)
        return True
    def ShowPlot(self):
        # Definition to plot the result.
        self.plt.show()
        return True

class raytracing():
    def __init__(self):
        # See "General Ray tracing procedure" from G.H. Spencerand M.V.R.K Murty for the theoratical explanation
        seterr(divide='ignore', invalid='ignore')
        pass
    def PlotInit(self):
        # Definition to initiate plot.
        self.plt = matplotlib.pyplot
        # New figure created.
        self.fig = self.plt.figure(figsize=(17, 13))
        # 3D projection is enabled.
        self.ax  = self.fig.gca(projection='3d')
        # Enabling the grid in the figure.
        self.ax.grid(True, color='k', linewidth=2)
        return True
    def SetPlotFontSize(self,family='normal',weight='normal',size='22'):
        # Definition to set the font type, size and weight in plots.
        font = {'family' : family,
                'weight' : weight,
                'size'   : size}
        matplotlib.rc('font', **font)
        # Enables Latex support in the texts.
        matplotlib.rc('text', usetex=True)
        # Set the ticks font propoerties as well to be on the safe side.
        self.ax.xaxis.label.set_fontsize(size)
        self.ax.yaxis.label.set_fontsize(size)
        self.ax.zaxis.label.set_fontsize(size)
        self.ax.xaxis.label.set_family(family)
        self.ax.yaxis.label.set_family(family)
        self.ax.zaxis.label.set_family(family)
        self.ax.xaxis.label.set_fontweight(weight)
        self.ax.yaxis.label.set_fontweight(weight)
        self.ax.zaxis.label.set_fontweight(weight)
        return True
    def DegreesToRadians(self,angle):
        # Function to convert degrees to radians.
        return radians(angle)
    def findangles(self,Point1,Point2):
        # Function to find angle between two points if there was a line intersecting at both of them.
        # Vector to hold angles is created.
        angles   = []
        # Find the distance in between points.
        distance = self.finddistancebetweentwopoints(Point1,Point2)
        # X axis rotation is calculated.
        angles.append(degrees(arccos( (Point2[0]-Point1[0])/distance )))
        # Y axis rotation is calculated.
        angles.append(degrees(arccos( (Point2[1]-Point1[1])/distance )))
        # Z axis rotation is calculated.
        angles.append(degrees(arccos( (Point2[2]-Point1[2])/distance )))
        # Angles are returned.
        return angles
    def GenerateBSpline(self, cv, n=100, degree=3, periodic=False):
        # Definition to generate a bspline.
        # Taken from http://stackoverflow.com/questions/34803197/fast-b-spline-algorithm-with-numpy-scipy
        # CV: control points.
        # n: number of samples.
        # degree: curve degree.
        # periodic: True - curve closed, False - curve open.
        n     = n**2
        # If periodic, extend the point array by count+degree+1
        cv    = asarray(cv)
        count = len(cv)
        if periodic:
            factor, fraction = divmod(count+degree+1, count)
            cv = np.concatenate((cv,) * factor + (cv[:fraction],))
            count = len(cv)
            degree = np.clip(degree,1,degree)
        # If opened, prevent degree from exceeding count-1
        else:
            degree = np.clip(degree,1,count-1)
        # Calculate knot vector
        kv = None
        if periodic:
           kv = np.arange(0-degree,count+degree+degree-1,dtype='int')
        else:
           kv = np.array([0]*degree + range(count-degree+1) + [count-degree]*degree,dtype='int')
        # Calculate query range
        u = np.linspace(periodic,(count-degree),n)
        # Calculate result
        arange = np.arange(len(u))
        points = np.zeros((len(u),cv.shape[1]))
        for i in range(cv.shape[1]):
            points[arange,i] = si.splev(u, (kv,cv[:,i],degree))
        return points
    def finddistancebetweentwopoints(self,Point1,Point2):
        # Function to find the distance between two points if there was a line intersecting at both of them.
        return sqrt(sum((array(Point1)-array(Point2))**2))
    def sphericaltocartesian(self,point,r,psi,teta):
        # Function to convert from spherical coordinates to cartesian coordinates.
        return array([
                      point[0]+r*sin(radians(psi))*cos(radians(teta)),
                      point[1]+r*sin(radians(psi))*sin(radians(teta)),
                      point[2]+r*cos(radians(teta))
                     ])
    def createvector(self,x0y0z0,abg):
        # Due to Python 2 -> Python 3.
        x0,y0,z0         = x0y0z0
        alpha,beta,gamma = abg
        # Create a vector with the given points and angles in each direction
        point = array([[x0],[y0],[z0]])
        alpha = cos(radians(alpha))
        beta  = cos(radians(beta))
        gamma = cos(radians(gamma))
        # Cosines vector.
        cosin = array([[alpha],[beta],[gamma]])
        return array([point.reshape(3,1),cosin.reshape(3,1)])
    def CalculateIntersectionOfTwoVectors(self,vector1,vector2):
        # Method to calculate the intersection of two vectors.
        A = array([
                   [vector1[1][0][0], vector2[1][0][0] ],
                   [vector1[1][1][0], vector2[1][1][0] ],
                   [vector1[1][2][0], vector2[1][2][0] ]
                  ])
        B = array([
                   vector1[0][0]-vector2[0][0],
                   vector1[0][1]-vector2[0][1],
                   vector1[0][2]-vector2[0][2]
                  ])
        # Least square solver.
        distances = scipy.linalg.lstsq(A,B)[0]
        # Check if the given solution matches the initial condition at the third equation.
        if allclose(dot(A,distances),B) == False:
           distances = [0,0]
        # Point vector created.
        Point = array([vector1[0][0][0]-distances[0]*vector1[1][0][0],
                       vector1[0][1][0]-distances[0]*vector1[1][1][0],
                       vector1[0][2][0]-distances[0]*vector1[1][2][0]
                      ])
        return Point,distances
    def createvectorfromtwopoints(self,x0y0z0,x1y1z1):
        # Because of Python 2 -> Python 3.
        x0,y0,z0 = x0y0z0
        x1,y1,z1 = x1y1z1
        # Create a vector from two given points.
        point = array([[x0],[y0],[z0]])
        # Distance between two points.
        s     = sqrt( pow((x0-x1),2) + pow((y0-y1),2) + pow((z0-z1),2) )
        if s != 0:
            alpha = (x1-x0)/s
            beta  = (y1-y0)/s
            gamma = (z1-z0)/s
        elif s == 0:
            alpha = float('nan')
            beta  = float('nan')
            gamma = float('nan')
        # Cosines vector
        cosin = array([[alpha],[beta],[gamma]])
        # returns vector and the distance.
        return array([point,cosin]),s
    def multiplytwovectors(self,vector1,vector2):
        # Multiply two vectors and return the resultant vector.
        # Used method described under:
        # Cross-product: http://en.wikipedia.org/wiki/Cross_product
        angle = cross(vector1[1].transpose()[0],vector2[1].transpose()[0])
        return array([vector1[0],[[angle[0]],[angle[1]],[angle[2]]]])
    def anglebetweentwovector(self,vector0,vector1):
        # Finds angle between two vectors
        # Used method described under: http://www.wikihow.com/Find-the-Angle-Between-Two-Vectors
        angle = vector0[1]*vector1[1]
        angle = angle[0]+angle[1]+angle[2]
        s1    = sqrt(vector0[1][0]**2+vector0[1][1]**2+vector0[1][2]**2)
        s2    = sqrt(vector1[1][0]**2+vector1[1][1]**2+vector1[1][2]**2)
        angle = degrees(arccos(angle/(s1*s2)))
        return angle
    def SameSide(self,p1,p2,a,b):
        # Methodology taken from http://www.blackpawn.com/texts/pointinpoly/
        ba    = subtract(b,a)
        p1a   = subtract(p1,a)
        p2a   = subtract(p2,a)
        cp1   = cross(ba,p1a)
        cp2   = cross(ba,p2a)
        if dot(cp1,cp2) >= 0:
            return True
        return False
    def isitontriangle(self,pointtocheck,point0,point1,point2,error=0.001):
        # Check if the given point is insight the triangle which represented.
        # point0, point1 and point2 are the corners of the triangle.
        pointtocheck = pointtocheck.reshape(3)
        side0        = self.SameSide(pointtocheck,point0,point1,point2)
        side1        = self.SameSide(pointtocheck,point1,point0,point2)
        side2        = self.SameSide(pointtocheck,point2,point0,point1)
        if side0 == True and side1 == True and side2 == True:
            return True
        return False
#    def transform(self,input,(alpha,beta,gamma),(x0,y0,z0)):
    def transform(self,input,abg,x0y0z0):
        # Because of Python 2 -> Python 3.
        alpha,beta,gamma = abg
        x0,y0,z0         = x0y0z0
        # alpha; rotation angle (euler) of x axis.
        # beta; rotation angle (euler) of y axis.
        # gamma; rotation angle (euler) of z axis.
        # x0; x coordinate of origin measured in the reference system.
        # y0; y coordinate of origin measured in the reference system.
        # z0; z coordinate of origin measured in the reference system.
        alpha  = radians(alpha)
        beta   = radians(beta)
        gamma  = radians(gamma)
        R1     = array([[cos(gamma),-sin(gamma),0],[sin(gamma),cos(gamma),0],[0,0,1]])
        R2     = array([[1,0,0],[0,cos(beta),-sin(beta)],[0,sin(beta),cos(beta)]])
        R3     = array([[cos(alpha),0,-sin(alpha)],[0,1,0],[sin(alpha),0,cos(alpha)]])
        R      = dot(dot(R1,R2),R3)
        output = dot(R,input-array([[x0],[y0],[z0]]))
        return output
    def reflect(self,vector,normvector):
        # Used method described in G.H. Spencer and M.V.R.K. Murty, "General Ray-Tracing Procedure", 1961
        mu          = 1
        div         = pow(normvector[1,0],2)  + pow(normvector[1,1],2) + pow(normvector[1,2],2)
        a           = mu* ( vector[1,0]*normvector[1,0]
                          + vector[1,1]*normvector[1,1]
                          + vector[1,2]*normvector[1,2]) / div
        outvec      = vector.copy()
        outvec[0,0] = normvector[0,0]
        outvec[0,1] = normvector[0,1]
        outvec[0,2] = normvector[0,2]
        outvec[1,0] = vector[1,0] - 2*a*normvector[1,0]
        outvec[1,1] = vector[1,1] - 2*a*normvector[1,1]
        outvec[1,2] = vector[1,2] - 2*a*normvector[1,2]
        return outvec
    def FindReflectNormal(self,vector0,vector1):
        # Definition to find reflection normal in between two given vectors.
        mu = 1
        NormVector      = vector1.copy()
        for k in range(1,3):
            a0              = arccos(vector0[1,k])
            a1              = arccos(vector1[1,k])
            NormVector[1,k] = cos(a1+(pi-(a0+a1))/2.)
        return NormVector
    def snell(self,vector,normvector,n1,n2,error=0.01):
        # Method for Snell's law
        # n1 refractive index of the medium which vector is coming from
        # n2 refractive index of the medium which vector tries to go into
        mu    = n1/n2
        div   = pow(normvector[1,0],2)  + pow(normvector[1,1],2) + pow(normvector[1,2],2)
        a     = mu* (vector[1,0]*normvector[1,0] + vector[1,1]*normvector[1,1] + vector[1,2]*normvector[1,2]) / div
        b     = (pow(mu,2)-1) / div
        to    = -b*0.5/a
        num   = 0
        eps   = error*2
        # Newton-Raphson method to find the correct root
        while eps > error:
           num   += 1
           oldto  = to
           v      = pow(to,2) + 2*a*to + b
           deltav = 2*(to+a)
           to     = to - v /deltav
           eps    = abs(oldto-to)
           # Iteration notifier
           #print 'Iteration number: %s, Error: %s' % (num,error)
           # Iteration limiter
           if num > 5000:
              return False
        VectorOutput      = vector.copy()
        VectorOutput[0,0] = normvector[0,0]
        VectorOutput[0,1] = normvector[0,1]
        VectorOutput[0,2] = normvector[0,2]
        VectorOutput[1,0] = mu*vector[1,0] + to*normvector[1,0]
        VectorOutput[1,1] = mu*vector[1,1] + to*normvector[1,1]
        VectorOutput[1,2] = mu*vector[1,2] + to*normvector[1,2]
        return VectorOutput
    def FuncPoly2nd(self,k,l,m,poly):
        # Definition to return a 3D point position in a second order polynomial surface.
        # a = sqrt(x^2+y^2)
        # k2 x (a)^2 + k1 x a + k0 a - z =0
        a = sqrt((k-poly[0])**2+(l-poly[1])**2)
        return poly[5]*a**2+poly[4]*a+poly[3]-m+poly[2]
    def FuncEllipsoid(self,k,l,m,ellipsoid):
        # Definition to return a 3D point position in quadratic surface definition.
        return pow((k-ellipsoid[0])/ellipsoid[4],2) + pow((l-ellipsoid[1])/ellipsoid[5],2) + pow((m-ellipsoid[2])/ellipsoid[6],2) - pow(ellipsoid[3],2)
    def FuncNormPoly2nd(self,x0,y0,z0,poly2nd):
        # Definition to return normal of a sphere.
        # Derivatives.
        # m = x^2+y^2
        # a = sqrt(m)
        m     = (x0-poly2nd[0])**2+(y0-poly2nd[1])**2
        a     = sqrt(m)
        dadx  = 0.5*1./a*(2*x0-2*poly2nd[0])
        dady  = 0.5*1./a*(2*y0-2*poly2nd[1])
        dzda  = 2*poly2nd[5]*a + poly2nd[4]
        gradx = dzda*dadx
        grady = dzda*dady
        gradz = -1
        # Perpendicular to tangent surface.
        alpha = degrees(arctan(gradx))+90; beta = degrees(arctan(grady))+90; gamma = degrees(arctan(gradz))+90
        # Return a normal vector.
        return self.createvector((x0,y0,z0),(alpha,beta,gamma))
    def FuncSpher(self,k,l,m,sphere):
        # Definition to return a 3D point position in spherical definition.
        return pow(k-sphere[0],2) + pow(l-sphere[1],2) + pow(m-sphere[2],2) - pow(sphere[3],2)
    def FuncNormSpher(self,x0,y0,z0,sphere):
        # Definition to return normal of a sphere.
        # Derivatives.
        gradx = 2*(x0-sphere[0])/sphere[3]**2
        grady = 2*(y0-sphere[1])/sphere[3]**2
        gradz = 2*(z0-sphere[2])/sphere[3]**2
        # Perpendicular to tangent surface.
        alpha = degrees(arctan(gradx))+90; beta = degrees(arctan(grady))+90; gamma = degrees(arctan(gradz))+90
        # Return a normal vector.
        return self.createvector((x0,y0,z0),(alpha,beta,gamma))
    def FuncNormEllipsoid(self,x0,y0,z0,ellipsoid):
        # Definition to return normal of a quadratic surface.
        # Derivatives.
        gradx = 2/ellipsoid[4]*(x0-ellipsoid[0])/ellipsoid[3]**2
        grady = 2/ellipsoid[5]*(y0-ellipsoid[1])/ellipsoid[3]**2
        gradz = 2/ellipsoid[6]*(z0-ellipsoid[2])/ellipsoid[3]**2
        # Perpendicular to tangent surface.
        alpha = degrees(arctan(gradx))+90; beta = degrees(arctan(grady))+90; gamma = degrees(arctan(gradz))+90
        # Return a normal vector.
        return self.createvector((x0,y0,z0),(alpha,beta,gamma))
    def FindInterPoly2nd(self,vector,poly2nd,error=0.00000001,numiter=1000,iternotify='no'):
        # Definition to return intersection of a ray with a quadratic surface.
        return self.FindInterFunc(vector,poly2nd,self.FuncPoly2nd,self.FuncNormPoly2nd,error=error,numiter=numiter,iternotify=iternotify)
    def FindInterEllipsoid(self,vector,ellipsoid,error=0.00000001,numiter=1000,iternotify='no'):
        # Definition to return intersection of a ray with a quadratic surface.
        return self.FindInterFunc(vector,ellipsoid,self.FuncEllipsoid,self.FuncNormEllipsoid,error=error,numiter=numiter,iternotify=iternotify)
    def findinterspher(self,vector,sphere,init=None,error=0.00000001,numiter=1000,iternotify='no'):
        # Definition to return intersection of a ray with a sphere.
        return self.FindInterFunc(vector,sphere,self.FuncSpher,self.FuncNormSpher,init=init,error=error,numiter=numiter,iternotify=iternotify)
    def FindInterFunc(self,vector,SurfParam,Func,FuncNorm,init=None,error=0.00000001,numiter=1000,iternotify='no'):
        # Method for finding intersection in between a vector and a parametric surface
        number   = 0
        distance = 1
        olddist  = 0
        shift = 0
        if init == None:
            distance = 1.
        else:
            distance = init
        epsilon  = error*2
        k        = vector[0,0,0]
        l        = vector[0,1,0]
        m        = vector[0,2,0]
        FXYZ     = Func(k,l,m,SurfParam)
        if abs(FXYZ) < 0.01:
            shift  = 1.5 * SurfParam[3]
            k     += shift * vector[1,0]
            l     += shift * vector[1,1]
            m     += shift * vector[1,2]
        while epsilon > error:
            number   += 1
            oldFXYZ   = FXYZ
            x         = distance * vector[1,0] + k
            y         = distance * vector[1,1] + l
            z         = distance * vector[1,2] + m
            FXYZ      = Func(x,y,z,SurfParam)
            # Secant method is calculated, see wikipedia article of the method for more
            newdist   = distance - FXYZ*(distance-olddist)/(FXYZ-oldFXYZ)
            epsilon   = abs(newdist-distance)
            olddist   = distance
            distance  = newdist
            # Check if the number of iterations are too much
            if number > numiter:
               return False,False
        normvec = FuncNorm(x,y,z,SurfParam)
        return distance+shift,normvec
    def findintersurface(self,vector,points):
        point0,point1,point2 = points
        # Method to find intersection point inbetween a surface and a vector
        # See http://geomalgorithms.com/a06-_intersect-2.html
        vector0,s     = self.createvectorfromtwopoints(point0,point1)
        vector1,s     = self.createvectorfromtwopoints(point1,point2)
        vector2,s     = self.createvectorfromtwopoints(point0,point2)
        normvec       = self.multiplytwovectors(vector0,vector2)
        f             = point0-vector[0].T
        n             = normvec[1].copy()
        distance      = dot(n.T,f.T)/dot(n.T,vector[1])
        normvec[0][0] = vector[0][0]+distance*vector[1][0]
        normvec[0][1] = vector[0][1]+distance*vector[1][1]
        normvec[0][2] = vector[0][2]+distance*vector[1][2]
        return distance[0][0],normvec
    def propagateray(self,vector,distance):
        # Method to propagate a ray.
        return array([
                      vector[0][0]+distance*vector[1][0],
                      vector[0][1]+distance*vector[1][1],
                      vector[0][2]+distance*vector[1][2]
                     ])
    def plotvector(self,vector,distance,color='g'):
        # Method to plot rays
        x = array([vector[0,0,0], distance * vector[1,0] + vector[0,0,0]])
        y = array([vector[0,1,0], distance * vector[1,1] + vector[0,1,0]])
        z = array([vector[0,2,0], distance * vector[1,2] + vector[0,2,0]])
        self.ax.plot(x,y,z,color)
        return True
    def PlotPoint(self,point,color='g*',contour=False,marker=False):
        # Method to plot a single spot.
        self.ax.plot(array([point[0]]),array([point[1]]),array([point[2]]),color)
        # Plotting contour on 3-axes.
        if contour == True:
            self.ax.contour(array([[-1,0,1],
                                   [-1,0,1],
                                   [-1,0,1]
                                   ]),
                            array([[-1,0,1],
                                   [-1,0,1],
                                   [-1,0,1]
                                   ]),
                            array([[2,4,2],
                                   [1,4,5],
                                   [3,3,3]
                                   ]),
                            zdir='z', offset=-100, cmap=cm.coolwarm)
        # Add text near to the point.
        if marker == True:
            label = '(%.1f, %.1f, %.1f)' % (point[0], point[1], point[2])
            self.ax.text(point[0], point[1], point[2], label)
        return True
    def plotsphericallens(self,cx=0,cy=0,cz=0,r=10,c='none',a=0.3,PlotFlag=True):
        # Method to plot a spherical lens.
        sampleno = 100
        v        = linspace(0, pi, sampleno)
        u        = linspace(0,2*pi,sampleno)
        x        = r * outer(cos(u), sin(v)) + cx
        y        = r * outer(sin(u), sin(v)) + cy
        z        = r * outer(ones(size(u)), cos(v)) + cz
        if PlotFlag == True:
            self.ax.plot_surface(x, y, z, rstride=8, cstride=8, alpha=a, color=c, antialiased=True)
        return array([cx,cy,cz,r])
    def CalculateParaboloidMesh(self,spher,sampleno=100,angleu=[0,pi],anglev=[0,2*pi]):
        # Definition to calculate triangular meshed form of a spherical sufrace.
        cx = spher[0]; cy = spher[1]; cz = spher[2]; r = spher[3]
        v           = linspace(angleu[0], angleu[1], sampleno)
        u           = linspace(anglev[0],anglev[1],sampleno)
        tris        = zeros((sampleno,sampleno,3))
        tris[:,:,0] = r * outer(cos(u), sin(v)) + cx
        tris[:,:,1] = r * outer(sin(u), sin(v)) + cy
        tris[:,:,2] = r * outer(ones(size(u)), cos(v)) + cz
        return tris
    def CalculateSpherMesh(self,spher,sampleno=100,angleu=[0,pi],anglev=[0,2*pi]):
        # Definition to calculate triangular meshed form of a spherical sufrace.
        cx = spher[0]; cy = spher[1]; cz = spher[2]; r = spher[3]
        v           = linspace(angleu[0], angleu[1], sampleno)
        u           = linspace(anglev[0],anglev[1],sampleno)
        tris        = zeros((sampleno,sampleno,3))
        tris[:,:,0] = r * outer(cos(u), sin(v)) + cx
        tris[:,:,1] = r * outer(sin(u), sin(v)) + cy
        tris[:,:,2] = r * outer(ones(size(u)), cos(v)) + cz
        return tris
    def CalculateFocal(self,rx,ry,rz,n):
        # Method to calculate the focal length of the lens in different axes.
        for a in [rx,ry]:
            R         = (pow(a,2)+pow(rz,2))/(2*rz)
            LensMaker = (n-1)*(-2/R+(n-1)*rz*2/(n*R))
            f         = pow(LensMaker,-1)
            prompt('Focal length of the lens: %' % f)
        return True
    def PlotMesh(self,tris,alpha=0.3):
        # Definition to plot meshes using triangles.
        sampleno = tris.shape[0]
        for i in range(0,sampleno-1):
            for j in range(0,sampleno-1):
                self.plottriangle(tris[i,j],tris[i+1,j],tris[i,j+1],alpha=alpha)
                self.plottriangle(tris[i+1,j+1],tris[i+1,j],tris[i,j+1],alpha=alpha)
        return tris
    def FindInterMesh(self,vector,tris,RangeWindow=None,s=2):
        # Definition to find the first intersection of a ray with a mesh.
        sampleno = tris.shape[0]
        if RangeWindow == None:
            limin = 0; limax=sampleno-1
            ljmin = 0; ljmax=sampleno-1
        else:
            limin = RangeWindow[0][0]; limax = RangeWindow[0][1]
            ljmin = RangeWindow[1][0]; ljmax = RangeWindow[1][1]
        for i in range(limin,limax):
            for j in range(ljmin,ljmax):
                tri0        = [tris[i,j],tris[i+1,j],tris[i,j+1]]
                tri1        = [tris[i+1,j+1],tris[i+1,j],tris[i,j+1]]
                s0,normvec0 = self.findintersurface(vector,(tri0[0],tri0[1],tri0[2]))
                res0        = self.isitontriangle(normvec0[0],tri0[0],tri0[1],tri0[2])
                if res0 == True:
                    return s0,normvec0,tri0,[[i-s,i+s],[j-s,j+s]]
                s1,normvec1 = self.findintersurface(vector,(tri1[0],tri1[1],tri1[2]))
                res1        = self.isitontriangle(normvec1[0],tri1[0],tri1[1],tri1[2])
                if res1 == True:
                    return s1,normvec1,tri1,[[i-s,i+s],[j-s,j+s]]
        return 0,None,None,None
    def PlotCircle(self,center,r,c='none'):
        # Method to plot circle.
        circle = Circle((center[0], center[1]), r, facecolor=c, edgecolor=(0,0,0), linewidth=4, alpha=1)
        self.ax.add_patch(circle)
        art3d.pathpatch_2d_to_3d(circle, z=center[2], zdir='z')
        return array([center,r])
    def PlotData(self,X,Y,Z,c='none'):
        # Method to plot the given data.
        # Gridding the data.
        xi = linspace(min(X), max(X))
        yi = linspace(min(Y), max(Y))
        xim, yim = meshgrid(xi, yi)
        zi = griddata(X,Y,Z,xi,yi,interp='nn')
        # Plot the resultant figure.
        self.ax  = self.fig.gca()
        self.ax.plot_surface(xim, yim, zi, rstride=2, cstride=2, cmap=cm.jet, alpha=0.3, color=c)
        return True
    def plottriangle(self,point0,point1,point2,alpha=0.3):
        # Method to plot triangular surface
        x     = array([ point0[0], point1[0], point2[0]])
        y     = array([ point0[1], point1[1], point2[1]])
        z     = array([ point0[2], point1[2], point2[2]])
        verts = [zip(x, y,z)]
        self.ax.add_collection3d(Poly3DCollection(verts,alpha=alpha))
        return array([point0,point1,point2])
    def plotcornercube(self,centerx,centery,centerz,pitch,revert=False):
        # Method to plot a single cornercube
        point00 = array([ centerx, centery, centerz])
        point10 = array([ centerx, centery, centerz])
        point20 = array([ centerx, centery, centerz])
        if revert == False:
            point01 = array([ centerx, centery-2*pitch/3, centerz+sqrt(2)*pitch/3 ])
            point11 = array([ centerx, centery-2*pitch/3, centerz+sqrt(2)*pitch/3 ])
            point21 = array([ centerx-pitch/sqrt(3), centery+pitch/3, centerz+sqrt(2)*pitch/3 ])
            point02 = array([ centerx-pitch/sqrt(3), centery+pitch/3, centerz+sqrt(2)*pitch/3 ])
            point12 = array([ centerx+pitch/sqrt(3), centery+pitch/3, centerz+sqrt(2)*pitch/3 ])
            point22 = array([ centerx+pitch/sqrt(3), centery+pitch/3, centerz+sqrt(2)*pitch/3 ])
        else:
            point01 = array([ centerx, centery+2*pitch/3, centerz+sqrt(2)*pitch/3 ])
            point11 = array([ centerx, centery+2*pitch/3, centerz+sqrt(2)*pitch/3 ]) 
            point21 = array([ centerx+pitch/sqrt(3), centery-pitch/3, centerz+sqrt(2)*pitch/3 ])
            point02 = array([ centerx+pitch/sqrt(3), centery-pitch/3, centerz+sqrt(2)*pitch/3 ])
            point12 = array([ centerx-pitch/sqrt(3), centery-pitch/3, centerz+sqrt(2)*pitch/3 ])
            point22 = array([ centerx-pitch/sqrt(3), centery-pitch/3, centerz+sqrt(2)*pitch/3 ])
        self.plottriangle(point00,point01,point02)
        self.plottriangle(point10,point11,point12)
        self.plottriangle(point20,point21,point22)
        return array([point00,point01,point02]),array([point10,point11,point12]),array([point20,point21,point22])
    def defineplotshape(self,xminxmax,yminymax,zminzmax):
        # Because of Python 2 -> Python 3.
        xmin,xmax = xminxmax
        ymin,ymax = yminymax
        zmin,zmax = zminzmax
        # Method to define plot shape.
        self.ax.set_xlim3d(xmin,xmax)
        self.ax.set_ylim3d(ymin,ymax)
        self.ax.set_zlim3d(zmin,zmax)
        return True
    def SavePlot(self,filename):
        # Definition to save the plotted figure. One should call it after showplot.
        self.plt.savefig(filename,bbox_inches='tight')
        return True
    def showplot(self,title=None,LabelX=None,LabelY=None,filename=None):
        # Shows the prepared plot
        if title != None or LabelX != None or LabelY!= None:
            self.plt.title(title)
            self.plt.xlabel(LabelX)
            self.plt.ylabel(LabelY)
        self.ax.view_init(10.0, -135.0)
        if filename != None:
            self.SavePlot(filename)
        self.plt.show()
        self.CloseFigure()
        return True
    def CloseFigure(self):
        # Method to close the last figure.
        self.plt.close()
        return True
    def CloseAllFigures(self):
        # Method to close all figures.
        self.plt.close('all')
        return True

class jonescalculus():
    def __init__(self):
        return
    def linearpolarizer(self,input,rotation=0):
        # Linear polarizer, rotation is in degrees and it is counter clockwise
        rotation        = radians(rotation)
        rotmat          = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        linearpolarizer = array([[1,0],[0,0]])
        linearpolarizer = dot(rotmat.transpose(),dot(linearpolarizer,rotmat))
        return dot(linearpolarizer,input)
    def circullarpolarizer(self,input,type='lefthanded'):
        # Circullar polarizer
        if type == 'lefthanded':
            circullarpolarizer = array([[0.5,-0.5j],[0.5j,0.5]])
        if type == 'righthanded':
            circullarpolarizer = array([[0.5,0.5j],[-0.5j,0.5]])
        return dot(circullarpolarizer,input)
    def quarterwaveplate(self,input,rotation=0):
        # Quarter wave plate, type determines the placing of the fast axis
        rotation = radians(rotation)
        rotmat   = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        qwp = array([[1,0],[0,-1j]])
        qwp = dot(rotmat.transpose(),dot(qwp,rotmat))
        return dot(qwp,input)
    def halfwaveplate(self,input,rotation=0):
        # Half wave plate
        rotation = radians(rotation)
        rotmat   = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        hwp      = array([[1,0],[0,-1]])
        hwp      = dot(rotmat.transpose(),dot(hwp,rotmat))
        return dot(hwp,input)
    def waveplate(self,input,retardation=0,rotation=0):
        # Waveplate with a given retardation angle.
        rotation    = radians(rotation)
        retardation = 2*radians(retardation)
        rotmat      = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        wp          = array([[1,0],[0,exp(-1j*retardation)]])
        wp          = dot(rotmat.transpose(),dot(wp,rotmat))
        return dot(wp,input)
    def birefringentplate(self,input,nx,ny,d,wavelength,rotation=0):
        # Birefringent plate, d thickness, nx and ny refractive indices
        rotation = radians(rotation)
        rotmat   = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        delta    = 2*pi*(nx-ny)*d/wavelength
        bfp      = array([[1,0],[0,exp(-1j*delta)]])
        bfp      = dot(rotmat.transpose(),dot(bfp,rotmat))
        return dot(bfp,input)
    def LCD(self,input,alpha,ne,n0,d,wavelength,rotation=0):
        # THIS DEFINITION IS DEFECTED, DO NOT USE IT!
        # Nematic liquid crystal, d cell thickness, extraordinary refrative index ne, ordinary refractive index n0,
        # alpha helical twist per meter in right-hand sense along the direction of wave propagation
        # alpha is calculated is by dividing cell thickness to 1 meter length
        # alpha    =  1 /d
        #http://books.google.com/books?id=0XhtwBpMtA8C&lpg=PA208&ots=t8NlS2vuDP&dq=TN%20LC%20jones%20matrix&hl=tr&pg=PA209#v=onepage&q=TN%20LC%20jones%20matrix&f=false
        rotation = radians(rotation)
        rotmat   = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        gamma    = 2*pi*(ne-n0)/wavelength*d
        desangle = 90.
        N        = desangle/gamma
        prompt('%s %s' % (gamma, N))
        lrot     = array([[cos(alpha*d),-sin(alpha*d)],[sin(alpha*d),cos(alpha*d)]])
        lretard  = array([[exp(-1j*gamma/2/N),0],[0,exp(1j*gamma/2/N)]])
#        llcd     = 
        lc       = dot(lrot,lcd)
        lc       = dot(rotmat.transpose(),dot(lc,rotmat))
        return dot(lc,input)
    def ferroliquidcrystal(self,input,tetat,ne,n0,d,wavelength,fieldsign='+',rotation=0):
        # Ferroelectric liquid crystal, d cell thickness, extraordinary refrative index ne, ordinary refractive index n0
        # Applied field sign determines the rotation angle
        rotation = radians(rotation)
        tetat    = radians(tetat)
        rotmat   = array([[cos(rotation),sin(rotation)],[-sin(rotation),cos(rotation)]])
        beta     = 2*pi*(ne-n0)/wavelength
        lrot1    = array([[cos(tetat),-sin(tetat)],[sin(tetat),cos(tetat)]])
        lrot2    = array([[cos(tetat),sin(tetat)],[-sin(tetat),cos(tetat)]])
        lretard  = array([[1,0],[0,exp(-1j*beta*d)]])
        if fieldsign == '+':
            lc = dot(dot(lrot1,lretard),lrot2)
        elif fieldsign == '-':
            lc = dot(dot(lrot2,lretard),lrot1)
        lc       = dot(rotmat.transpose(),dot(lc,rotmat))
        return dot(lc,input)
    def electricfield(self,a1,a2):
        # Electric field vector is defined here.
        # a1 is the electic field intensity at x-axis.
        # a2 is the electric field intensity at y-axis.
        return array([[a1],[a2]])

class aperture():
    def __init__(self):
        self.plt = matplotlib.pyplot
        return
    def twoslits(self,nx,ny,X,Y,delta):
        # Creates a matrix that contains two slits
        obj=zeros((nx,ny),dtype=complex)
        for i in range(int(nx/2-X/2),int(nx/2+X/2)):
            for j in range(int(ny/2+delta/2-Y/2),int(ny/2+delta/2+Y/2)):
                obj[ny/2-abs(ny/2-j),i] = 1
                obj[j,i] = 1
        return obj
    def rectangle(self,nx,ny,side):
        # Creates a matrix that contains rectangle
        obj=zeros((nx,ny),dtype=complex)
        for i in range(int(nx/2-side/2),int(nx/2+side/2)):
            for j in range(int(ny/2-side/2),int(ny/2+side/2)):
                obj[j,i] = 1
        return obj
    def circle(self,nx,ny,radius):
        # Creates a matrix that contains circle
        obj=zeros((nx,ny),dtype=complex)
        for i in range(int(nx/2-radius/2),int(nx/2+radius/2)):
            for j in range(int(ny/2-radius/2),int(ny/2+radius/2)):
                if (abs(i-nx/2)**2+abs(j-ny/2)**2)**(0.5)< radius/2:
                    obj[j,i] = 1
        return obj
    def sinamgrating(self,nx,ny,grating):
        # Creates a sinuosidal grating matrix
        obj=zeros((nx,ny),dtype=complex)
        for i in range(nx):
            for j in range(ny):
                obj[i,j] = 0.5+0.5*cos(2*pi*j/grating)
        return obj
    def lens(self,nx,ny,focal,wavelength,pixeltom):
        # Creates a lens matrix
        obj    = zeros((nx,ny),dtype=complex)
        k      = 2*pi/wavelength
        x      = linspace(-nx*pixeltom,nx*pixeltom,nx)
        y      = linspace(-ny*pixeltom,ny*pixeltom,ny)
        X,Y    = meshgrid(x,y)
        Z      = X**2+Y**2
        obj    = exp(-1j*k*0.5/focal*Z)
        return obj
    def gaussian(self,nx,ny,sigma):
        # Creates a 2D gaussian matrix
        obj = zeros((nx,ny),dtype=complex)
        for i in range(nx):
            for j in range(ny):
                obj[i,j] = 1/pi/pow(sigma,2)*exp(-float(pow(i-nx/2,2)+pow(j-ny/2,2))/2/pow(sigma,2))
        return obj
    def retroreflector(self,nx,ny,wavelength,pitch,type='normal'):
        if nx != ny:
           nx = max([nx,ny])
           ny = nx
        part  = zeros((pitch,int(pitch/2)))
        for i in range(int(sqrt(3)*pitch/6)):
            for j in range(int(pitch/2)):
                if float(j)/(int(sqrt(3)*pitch/6)-i) < sqrt(3):
                    part[i,j]       = int(sqrt(3)*pitch/6)-i
                    part[pitch-i-1,int(pitch/2)-j-1] = part[i,j]
        for i in range(int(pitch)):
            for j in range(int(pitch/2)):
                if j != 0:
                    if float(j)/(int(pitch)-i) < 0.5 and (int(sqrt(3)*pitch/6)-i)/float(j) < 1./sqrt(3):
                        # Distance to plane determines the level of the amplitude
                        # Plane as a line y = slope*x+ pitch
                        # Perpendicula line  y = -(1/slope)*x+n
                        slope     = -0.5
                        n         = j + (1/slope) * (i)
                        x1        = (n - pitch/2)/(slope+1/slope)
                        y1        = -(1/slope)*x1+n
                        part[i,j] = int(sqrt(3)*pitch/6) - sqrt( pow(i-x1,2) + pow(j-y1,2) )
                        part[pitch-i-1,int(pitch/2)-j-1] = part[i,j]
                else:
                    if i > int(sqrt(3)*pitch/6):
                        slope     = -0.5
                        n         = j + (1/slope) * (i)
                        x1        = (n - pitch/2)/(slope+1/slope)
                        y1        = -(1/slope)*x1+n
                        part[i,j] = int(sqrt(3)*pitch/6) - sqrt( pow(i-x1,2) + pow(j-y1,2) )
                        part[pitch-i-1,int(pitch/2)-j-1] = part[i,j]
        left  = part
        right = part[::-1]
        part  = append(left,right,axis=1)
        obj   = tile(part,(nx/pitch,ny/pitch))
        for i in range(nx/pitch/2):
           obj[(2*i+1)*pitch:(2*i+1)*pitch+pitch,:] = roll(obj[(2*i+1)*pitch:(2*i+1)*pitch+pitch,:],pitch/2)
        k     = 2*pi/wavelength
        D     = 5
        obj   = pow(obj,3)*exp(1j*k*obj)
        return obj
    def show(self,obj,pixeltom,wavelength,title='Detector',type='normal',filename=None,xlabel=None,ylabel=None):
        # Plots a detector showing the given object
        self.plt.figure(),self.plt.title(title)
        nx,ny = obj.shape
        # Number of the ticks to be shown both in x and y axes
        if type == 'normal':
            obj = abs(obj)
        elif type == 'log':
            obj = log(abs(obj))
        img = self.plt.imshow(obj,cmap=matplotlib.cm.jet,origin='lower')
        self.plt.colorbar(img,orientation='vertical')
        self.plt.xlabel(xlabel)
        self.plt.ylabel(ylabel)
        if filename != None:
            self.plt.savefig(filename)
        return True
    def show3d(self,obj):
        nx,ny   = obj.shape
        fig     = self.plt.figure()
        ax      = fig.gca(projection='3d')
        X,Y     = mgrid[0:nx,0:ny]
        surf    = ax.plot_surface(X,Y,abs(obj), rstride=1, cstride=1, cmap=matplotlib.cm.jet,linewidth=0, antialiased=False)
        fig.colorbar(surf, shrink=0.5, aspect=5)
        return True
    def showrow(self,obj,wavelength,pixeltom,distance,filename=None):
        # Plots row crosssection of the given object
        nx,ny = obj.shape
        a     = 5
        self.plt.figure()
        self.plt.plot(arange(-nx/2,nx/2)*pixeltom,abs(obj[nx/2,:]))
        if filename != None:
            self.plt.savefig(filename)
        return True
    def showplots(self):
        # Definition to show the plots.
        self.plt.show()
        return True
    def ClosePlot(self):
        # Definition to kill a figure.
        self.plt.close("all")
        return True

class beams():
    def __init(self):
        self.plt = matplotlib.pyplot
        return
    def spherical(self,nx,ny,distance,wavelength,pixeltom,focal,amplitude=1):
        # Spherical wave
        distance = abs(focal-distance)
        k        = 2*pi/wavelength
        X,Y      = mgrid[-nx/2:nx/2,-ny/2:ny/2]*pixeltom
        r        = sqrt(pow(X,2)+pow(Y,2)+pow(distance,2))
        U        = amplitude/r*exp(-1j*k*r)
        return U
    def gaussian(self,nx,ny,distance,wavelength,pixeltom,amplitude,waistsize,focal=0):
        # Gaussian beam
        distance = abs(distance-focal)
        X,Y      = mgrid[-nx/2:nx/2,-ny/2:ny/2]*pixeltom
        ro       = sqrt(pow(X,2)+pow(Y,2))
        z0       = pow(waistsize,2)*pi/wavelength
        A0       = amplitude/(1j*z0)
        if distance == 0:
            U = A0*exp(-pow(ro,2)/pow(waistsize,2))
            return U
        k        = 2*pi/wavelength
        R        = distance*(1+pow(z0/distance,2))
        W        = waistsize*sqrt(1+pow(distance/z0,2))
        ksi      = 1./arctan(distance/z0)
        U        = A0*waistsize/W*exp(-pow(ro,2)/pow(W,2))*exp(-1j*k*distance-1j*pow(ro,2)/2/R+1j*ksi)
        return U
    def multiplication(self,wave,aperture):
        # Definition to multiply a wave with an aperture.
        return  multiply(wave,aperture)

class diffractions():
    def __init__(self):
        return
    def fft(self,obj):
        return fftshift(fft2(obj))
    def fresnelfraunhofer(self,wave,wavelength,distance,pixeltom,aperturesize,type='Fresnel'):
        # Definitions for Fresnel impulse respone (IR), Fresnel Transfer Function (TF), Fraunhofer diffraction.
        # According to "Computational Fourier Optics" by David Vuelz.
        nu,nv  = wave.shape
        k      = 2*pi/wavelength
        x      = linspace(-nu*pixeltom,nu*pixeltom,nu)
        y      = linspace(-nv*pixeltom,nv*pixeltom,nv)
        X,Y    = meshgrid(x,y)
        Z      = X**2+Y**2
        if type == 'Fresnel':
            smplng = wavelength*distance/(aperturesize*pixeltom)
            if pixeltom < smplng: # Fresnel Impulse Response
                h      = 1./(1j*wavelength*distance)*exp(1j*k*0.5/distance*Z)
                h      = fft2(fftshift(h))*pixeltom**2
                prompt('IR Fresnel method')
            else: # Fresnel Transfer Function
                h      = exp(1j*k*distance)*exp(-1j*pi*wavelength*distance*Z)
                h      = fftshift(h)
                prompt('TR Fresnel method')
            U1     = fft2(fftshift(wave))
            U2     = h*U1
            result = ifftshift(ifft2(U2))
        elif type == 'Fraunhofer':
            c      = 1./(1j*wavelength*distance)*exp(1j*k*0.5/distance*Z)
            result = c*ifftshift(fft2(fftshift(wave)))*pixeltom**2
        return result
    def SuggestType(self,aperturesize,pixeltom,wavelength,distance):
        # Definition to suggest, which type of propagation would be meaningful.
        fresnelno    = self.fresnelnumber(aperturesize,pixeltom,wavelength,distance)
        if fresnelno < 0.001:
            type = 'Fraunhofer'
        else:
            type = 'Fresnel'
        prompt('Fresnel number: %s, Propagation type: %s' % (fresnelno,type))
        return type,fresnelno
    def fresnelnumber(self,aperturesize,pixeltom,wavelength,distance):
        # Definition to calculate the fresnel number.
        return  pow(aperturesize*pixeltom,2)/wavelength/distance
    def intensity(self,obj):
        # Definition to calcualte the intensity of the given value.
        return abs(obj**2)
    def phase(self,obj):
        # Definition to show the phase of the given wave.
        return angle(obj)
    def transmittance(self,wave,trans):
        # Definition to calculate the output of a given input wave through a transmittance function.
        return multiply(wave,trans)
    def FWHM (self,Irow):
        # Definition to calculate full width at half maximum of a gaussian alike function.
        I  = Irow/sum(Irow) # Normalizing std
        return 2*sqrt(2*log(2))*std(I)

def main():
    prompt('Odak by %s' % __author__)
    return True

if __name__ == '__main__':
    sys.exit(main())




import sys
import os
import odak
from odak import np
from odak.wave import wavenumber,plane_tilt,prism_phase_function,quadratic_phase_function

def main():
    # Variables to be set.
    wavelength                 = 0.5*pow(10,-6)
    pixeltom                   = 6*pow(10,-6)
    distance                   = 10.0
    propagation_type           = 'Fraunhofer'
    k                          = wavenumber(wavelength)
    sample_field               = np.random.rand(150,150).astype(np.complex64)
    plane_field                = plane_tilt(
                                            sample_field.shape[0],
                                            sample_field.shape[1],
                                            k,
                                            [0.3,0.9,1.,1.],
                                            dx=pixeltom,
                                            axis='x'
                                           )
    lens_field                 = quadratic_phase_function(
                                                          sample_field.shape[0],
                                                          sample_field.shape[1],
                                                          k,
                                                          focal=0.3,
                                                          dx=pixeltom
                                                         )
    prism_field                = prism_phase_function(
                                                      sample_field.shape[0],
                                                      sample_field.shape[1],
                                                      k,
                                                      angle=0.1,
                                                      dx=pixeltom,
                                                      axis='x'
                                                     )

    sample_field               = sample_field*plane_field

    #from odak.visualize.plotly import detectorshow
    #detector       = detectorshow()
    #detector.add_field(sample_field)
    #detector.show()
    #detector       = detectorshow()
    #detector.add_field(hologram)
    #detector.show()
    #detector       = detectorshow()
    #detector.add_field(reconstruction)
    #detector.show()
    assert True==True

if __name__ == '__main__':
    sys.exit(main())



import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

thelibFolder = os.path.dirname(os.path.realpath(__file__))
requirementPath = thelibFolder + '/requirements.txt'
install_requires = []
if os.path.isfile(requirementPath):
    with open(requirementPath) as f:
        install_requires = f.read().splitlines()
        
setup(
      name                 = "odak",
      version              = "0.1.8",
      author               = "Kaan Akşit",
      author_email         = "kunguz@gmail.com",
      description          = (
                              "Odak (pronounced O-dac) is the fundamental Python library"
                              "for scientific computing in optical sciences."
                              "Odak includes modules for geometric raytracing and wave optics."
                             ),
      license              = "",
      keywords             = "optics holography",
      url                  = "https://github.com/kunguz/odak",
      install_requires     = install_requires,
      packages             = [
                              'odak',
                              'odak/raytracing',
                              'odak/jones',
                              'odak/tools',
                              'odak/wave',
                              'odak/learn/wave',
                              'odak/learn/tools',
                              'odak/visualize',
                              'odak/visualize/blender',
                              'odak/manager',
                              'odak/measurement',
                              'odak/catalog',
                              'odak/learn',
                              'odak/oldschool'
                             ],
      package_dir          = {'odak': 'odak'},
      package_data         = {'odak': ['catalog/data/*.json']},
      data_files           = [('', ['LICENSE.txt','README.md','THANKS.txt','requirements.txt'])],
      long_description     = read('README.md'),
      classifiers          = [
                              "Development Status :: 2 - Pre-Alpha",
                              "Intended Audience :: Science/Research",
                              "Intended Audience :: Developers",
                              "Topic :: Scientific/Engineering :: Physics",
                              "Programming Language :: Python",
                              "License :: OSI Approved :: Apache Software License",
                             ],
      python_requires      = '>=3.7.5',
     )



import odak.raytracing as raytracer
import sys
from odak.tools.sample import grid_sample,batch_of_rays

def test():
    sample_entry_points = grid_sample(
                                      no=[4,4],
                                      size=[100.,100.],
                                      center=[0.,0.,0.],
                                      angles=[0.,0.,0.]
                                     )
    sample_exit_points  = grid_sample(
                                      no=[4,4],
                                      size=[100.,100.],
                                      center=[0.,0.,100.],
                                      angles=[0.,0.,0.]
                                     )
    rays                = raytracer.create_ray_from_two_points(
                                                               sample_entry_points,
                                                               sample_exit_points
                                                              )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



#!/usr/bin/env python

import sys

def test_jones():
    import odak.jones as jones
    u_in  = jones.electricfield(0.5,0.)
    u_out = jones.linearpolarizer(u_in,rotation=0)
    print(u_out)
    assert True==True

if __name__ == '__main__':
    sys.exit(test_jones())



import odak.raytracing as raytracer
import odak.catalog as catalog
import sys
from odak.tools.sample import grid_sample,batch_of_rays

def test():
    sample_entry_points      = grid_sample(
                                           no=[2,2],
                                           size=[100.,100.],
                                           center=[0.,0.,0.],
                                           angles=[0.,0.,0.]
                                          )
    sample_exit_points       = grid_sample(
                                           no=[2,2],
                                           size=[100.,100.],
                                           center=[0.,50.,100.],
                                           angles=[0.,0.,0.]
                                          )
    rays                     = raytracer.create_ray_from_two_points(
                                                                    sample_entry_points,
                                                                    sample_exit_points
                                                                   )
    diffuser                 = catalog.diffusers.thin_diffuser(
                                                               shape=[50,50],
                                                               center=[0.,0.,100.],
                                                               angles=[0.,0.,0.],
                                                               diffusion_no=[5,2],
                                                               diffusion_angle=10.
                                                              )
    new_rays,normal,distance = diffuser.raytrace(rays)
    detector                 = catalog.detectors.plane_detector(
                                                                resolution=[5,5],
                                                                shape=[50,50],
                                                                center=[0.,0.,200.],
                                                                angles=[0.,0.,0.],
                                                               )
    normals,distances        = detector.raytrace(new_rays)
#    import odak
#    ray_visualize = odak.visualize.rayshow(line_width=5)
#    ray_visualize.add_line(rays[:,0],normal[:,0])
#    ray_visualize.add_line(new_rays[:,0],normals[:,0])
#    ray_visualize.show()
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



import odak.raytracing as raytracer
import sys
from odak import np

def test():
    start0       = np.array([0., 5.,0.])
    start1       = np.array([0.,-5.,0.])
    intersection = np.array([0.,0.,100.])
    ray0         = raytracer.create_ray_from_two_points(
                                                        start0,
                                                        intersection
                                                       )
    ray1         = raytracer.create_ray_from_two_points(
                                                        start1,
                                                        intersection
                                                       )
    c0,c1        = raytracer.find_nearest_points(ray0,ray1)
    print(c0,c1)
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



"""
``odak``
===================
Odak is the fundamental Python library needed for scientific computing in optical sciences. It includes modules for geometric raytracing and wave optics.

"""
try:
    import cupy as np
except:
    import numpy as np
import odak.tools
import odak.raytracing
import odak.wave
import odak.visualize
import odak.manager
import odak.jones
import odak.catalog
import odak.measurement

version_info = 0, 1, 8
__version__  = '.'.join(map(str, version_info))



#!/usr/bin/env python

import sys
import numpy as np
import odak
import odak.raytracing

def test_ray_intersect_w_a_surface():
    # Input points to create a ray.
    ray_start_point    = [5.,5.,0.]
    ray_end_point      = [0.,0.,1000.]
    # Create from two points.
    ray                = odak.raytracing.create_ray_from_two_points(
                                                                    ray_start_point,
                                                                    ray_end_point
                                                                   )
    # Intersection with the surface.
    surface_center     = [ 0., 0.,1000.]
    surface_angles     = [ 0., 0., 0.]
    surface_points     = odak.raytracing.define_plane(
                                                      surface_center,
                                                      surface_angles
                                                     )
    normal,distance    = odak.raytracing.intersect_w_surface(
                                                             ray,
                                                             surface_points
                                                            )
    assert True==True

if __name__ == '__main__':
    sys.exit(test_ray_intersect_w_a_surface())



"""
Wrappers for blender client calling definitions from `odak/visualize/blender/libblend.py` at the server side.
"""
import socket

host = 'localhost'
port = 8082

def send_msg(cmds):
    """
    Definition to send command to blender server started by `odak/visualize/blender/server.py`.Make sure that you call init() definition from `odak/visualize/blender/__init__.py` to start the server first.

    Parameters
    ----------
    cmds           : list
                     List of pythonic commands to be transmittted to the blender server.

    Returns
    ----------
    state          : bool
                     Returns True if command is successfully sent.
    """
    try:
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    except:
        return False
    client.connect((host,port))
    for cmd in cmds:
        client.sendall(cmd.encode('utf-8') + b'\x00')
    client.shutdown(0)
    client.close()
    return True

def import_ply(filename,location=[0,0,0],angle=[0,0,0],scale=[1,1,1]):
    """
    Definition to import a PLY object.

    Parameters
    ----------
    filename       : str
                     Path of the PLY file to be imported.
    location       : list
                     location of the imported object along X,Y, and Z axes.
    angle          : list
                     Euler angles for rotating the imported object along X,Y, and Z axes.
    scale          : list
                     Scale of the imported object.

    Returns
    ----------
    cmd             : str
                      Command sent by a client.
    """
    cmd = [
           'import_ply(filename="%s",location=%s,angle=%s,scale=%s)' % (filename,location,angle,scale),
          ]
    print(cmd)
    send_msg(cmd)
    return cmd

def prepare(resolution=[1920,1080],camera_fov=40.0,camera_location=[0.,0.,-15.],camera_lookat=[0.,0.,0.],clip=[0.1,100000000.],device='GPU',intensity=2.,world_color=[0.3,0.3,0.3]):
    """
    Definition to prepare the viewport in Blender.

    Parameters
    ----------
    resolution      : list
                      Resolution of the viewport.
    camera_fov      : float
                      Viewport field of view in angles.
    camera_location : list
                      Viewport location in XYZ.
    clip            : list
                      Clipping planes near and far.
    device          : str
                      Device to be used for rendering CPU or GPU.
    intensity       : float
                      World's illumination intensity.
    world_color     : list
                      RGB color of the world's illumination.
 
    Returns
    ----------
    cmd             : str
                      Command sent by a client.
    """ 
    cmd = [
           'prepare(resolution=%s,camera_fov=%s,camera_location=%s,camera_lookat=%s,clip=%s,device="%s",intensity=%s,world_color=%s)' % (resolution,camera_fov,camera_location,camera_lookat,clip,device,intensity,world_color),
          ]
    print(cmd)
    send_msg(cmd)
    return cmd

def delete_the_cube():
    """
    Definitioni to delete the default cube that appears every time one starts blender.

    Returns
    ----------
    cmd             : str
                      Command sent by a client.   Returns
    """
    cmd = [
           'delete_object("Cube")'
          ]
    print(cmd)
    send_msg(cmd)
    return cmd

def render(fn,exit=False):
    """
    Definition to render a scene, and save it as a PNG file.

    Parameters
    ----------
    fn             : str
                     Filename.

    Returns
    ----------
    cmd            : str
                     Command sent by a client.   
    exit           : bool                      
                     When set to True blender exits upon rendering completion.                      
    """
    cmd = [
           'render("%s",exit=%s)' % (fn,exit),
          ]
    print(cmd)
    send_msg(cmd)
    return cmd

def cylinder_between(start_loc,end_loc,r=0.1,objname='cylinder',color=[0.,0.5,0.,0.]):
    """ 
    Definition to create a cylinder in between two points with a certain radius.

    Parameters
    ----------
    stat_loc       : list
                     List that contains X,Y and Z start location.
    end_loc        : list
                     List that contains X,Y and Z end location.
    r              : float
                     Radius of the cylinder.
    objname        : str
                     Label of the object to be created.
    color          : list
                     Color of the cylinder (RGBA)

    Returns
    ----------
    obj            : blender object
                     Created cylinder.
    """
    cmd = [
           'cylinder_between(%s,%s,r=%s,objname="%s",color=%s)' %(start_loc,end_loc,r,objname,color)
          ]
    print(cmd)
    send_msg(cmd)
    return cmd


def quit():
    """
    Definition to quit blender.

    Returns
    ----------
    cmd             : str
                      Command sent by a client.   
    """
    cmd = [
           'quit',
          ]
    print(cmd)
    send_msg(cmd)
    return cmd



import odak.raytracing as raytracer
import sys
from odak import np

def detector_to_light_source(detector_location,triangle,light_source):
    center_of_triangle        = raytracer.center_of_triangle(triangle)
    ray_detector_triangle     = raytracer.create_ray_from_two_points(
                                                                     detector_location,
                                                                     center_of_triangle
                                                                    )
    normal_triangle,d_det_tri = raytracer.intersect_w_triangle(
                                                               ray_detector_triangle,
                                                               triangle
                                                              )
    if d_det_tri == 0:
        return 0
    ray_reflection            = raytracer.reflect(
                                                  ray_detector_triangle,
                                                  normal_triangle
                                                 )
    normal_source,d_tri_sou   = raytracer.intersect_w_circle(
                                                             ray_reflection,
                                                             light_source
                                                            )
    if d_tri_sou == 0:
        return 0
    opl                       = d_det_tri + d_tri_sou
    return opl

def test():
    detector_location = [2.,0.,0.]
    triangle          = np.array(
                                 [
                                  [ 10., 10., 10.],
                                  [  0., 10., 10.],
                                  [  0.,  0., 10.]
                                 ]
                                )
    circle_center     = [0.,0.,0.]
    circle_angles     = [0.,0.,0.]
    circle_radius     = 15.
    circle            = raytracer.define_circle(
                                                angles=circle_angles,
                                                center=circle_center,
                                                radius=circle_radius
                                               )

    opl               = detector_to_light_source(
                                                 detector_location,
                                                 triangle,
                                                 circle
                                                )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())
